// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*
   C interface definition for DAL

   (c) 1997 European Space Agency - written by Mark Thomas, CARA

*/


#ifndef __DAL_H
#define __DAL_H

#include "machine_specific_types.h"
#include "types.h"


typedef enum { Read = 1, Create, Modify, Temp, AsParent } AccessMode;
typedef enum { High = 1, HighLow, Low, UseEnvironment } MemoryModel;
typedef enum { TableType = 1, ArrayType } BlockType;
typedef enum { EraseAllFirst = 1, Merge } CopyMode;
typedef enum { Fixed = 1, Variable } CellType;
typedef enum { IntegerNull = 1, RealNull, StringNull, Undefined } NullType;
typedef enum { BooleanAttribute = 1, IntegerAttribute, RealAttribute, StringAttribute } AttributeType;
typedef enum { ArrayS = 1, AttributableS, AttributeS, BlockS, ColumnS, DataComponentS, DataSetS, LabelledS, NullableS, RowS, SubTableS, TableS } StructType;

typedef struct { StructType	structType; } Array;
typedef struct { StructType	structType; } Attributable;
typedef struct { StructType	structType; } Attribute;
typedef struct { StructType	structType; } Block;
typedef struct { StructType	structType; } Column;
typedef struct { StructType	structType; } DataComponent;
typedef struct { StructType	structType; } DataSet;
typedef struct { StructType	structType; } Labelled;
typedef struct { StructType	structType; } Nullable;
typedef struct { StructType	structType; } Row;
typedef struct { StructType	structType; } SubTable;
typedef struct { StructType	structType; } Table;
typedef void * TableIteratorFunction;	
/* typedef void(*TableIteratorFunction)( Table * ); */

#define MAX_LABEL_LENGTH 1024
#define MAX_NAME_LENGTH 16
#define MAX_UNITS_LENGTH 16

typedef char Label[ MAX_LABEL_LENGTH ];
typedef char Name[ MAX_NAME_LENGTH ];
typedef char Units[ MAX_UNITS_LENGTH ];

/* extern "C" { */

Attribute * addAttribute(
		Attributable*	attributable,
                const char*	name,
		const char*	comment,
                const char*	units );

void setBoolAttributeValue(
		Attribute*	attribute,
                bool8		value );

void setIntAttributeValue(
		Attribute*	attribute,
                int		value );

void setRealAttributeValue(
		Attribute*	attribute,
                double		value );

void setStringAttributeValue(
		Attribute*	attribute,
                const char*	value );

Attribute * addIntAttribute(
		Attributable*	attributable,
                const char*	name,
		int		value,
		const char*	comment,
                const char*	units );

Attribute * addRealAttribute(
		Attributable*	attributable,
                const char*	name,
		double		value,
		const char*	comment,
                const char*	units );

Attribute * addStringAttribute(
		Attributable*	attributable,
                const char*	name,
		const char*	value,
		const char*	comment );

Attribute * addBoolAttribute(
		Attributable*	attributable,
                const char*	name,
		bool8		value,
		const char*	comment );

Array * addArray(
		DataSet *			dataSet,
		const char *			name,
		DataType			dataType,
		int				numberofDimensions,
		unsigned long *			dimensions,
		const char *			units,
		const char *			label,
		int				position );

Column * addColumn(
		Table *				table,
		const char *			name,
		DataType			dataType,
		const char *			units,
		const char *			label,
		int				numberofDimensions,
		const unsigned long *		dimensions,
		int				position );

void addCommentToArray(
		Array *				array,
		const char *			comment );

void addCommentToAttributable(
		Attributable *			attributable,
		const char *			comment );

void addCommentToBlock(
		Block *				block,
		char *				comment );

void addCommentToDataSet(
		DataSet *			dataSet,
		char *				comment );

void addCommentToTable(
		Table *				table,
		char *				comment );

void addHistoryToArray(
		Array *				array,
		char *				text );

void addHistoryToAttributable(
		Attributable *			attributable,
		const char *			history );

void addHistoryToBlock(
		Block *				block,
		char *				text );

void addHistoryToDataSet(
		DataSet *			dataSet,
		char *				text );

void addHistoryToTable(
		Table *				table,
		char *				text );

Table * addTable(
		DataSet *			dataSet,
		char *				name,
		unsigned int 			numberOfRows,
		const char *			label,
		int 				position );

const char * arrayAttributeComment(
		const Array *			array,
		const char *			name );

AttributeType arrayAttributeDataTypeWithName(
		const Array *			array,
		const char *			name );

AttributeType arrayAttributeDataTypeWithNumber(
		const Array *			array,
		unsigned int			number );

const char * arrayAttributeUnits(
		const Array *			array,
		const char *			name );

Attribute * arrayAttributeWithName(
		const Array *			array,
		const char *			name );

Attribute * arrayAttributeWithNumber(
		const Array *			array,
		unsigned int			number );

DataType arrayDataType(
		const Array *			array );

bool8 arrayHasAttribute(
		const Array *			array,
		const char *			name );

const char * arrayLabel(
		const Array *			array );

Attributable * arrayToAttributable(
		Array *				array );

Block * arrayToBlock(
		Array *				array );

Nullable * arrayToNullable(
		Array *				array );

const char * arrayUnits(
		const Array *			array );

Array * arrayWithName(
		const DataSet *			dataSet,
		const char *			name );

Array * arrayWithNumber(
		const DataSet *			dataSet,
		unsigned int			position );

AttributeType attributableAttributeDataTypeWithName(
		const Attributable *		attributable,
		const char *			name );

AttributeType attributableAttributeDataTypeWithNumber(
		const Attributable *		attributable,
		unsigned int			number );

Attribute * attributableAttributeWithName(
		const Attributable *		attributable,
		const char *			name );

Attribute * attributableAttributeWithNumber(
		const Attributable *		attributable,
		unsigned int			number );

AttributeType attributeDataType(
		Attribute *			attribute );

const char * attributeLabel(
		const Attribute *		attribute );

const char * attributeUnits(
		const Attribute *		attribute );

const char * blockAttributeComment(
		const Block *			block,
		const char *			name );

AttributeType blockAttributeDataTypeWithName(
		const Block *			block,
		const char *			name );

AttributeType blockAttributeDataTypeWithNumber(
		const Block *			block,
		unsigned int			number );

const char * blockAttributeUnits(
		const Block *			dataSet,
		const char *			name );

Attribute * blockAttributeWithName(
		const Block *			block,
		const char *			name );

Attribute * blockAttributeWithNumber(
		const Block *			block,
		unsigned int			number );

bool8 blockHasAttribute(
		const Block *			block,
		const char *			name );

const char * blockLabel(
		const Block *			block );

unsigned int blockNumber(
		const DataSet *			dataSet,
		const char *			name );

Attributable * blockToAttributable(
		Block *				block );

BlockType blockType( const Block *		block );

Block * blockWithName(
		const DataSet *			dataSet,
		const char *			name );

Block * blockWithNumber(
		const DataSet *			dataSet,
		int				position );

bool8 boolArrayAttribute(
		const Array *			array,
		const char *			name );

bool8 * boolArrayData(
		const Array *			array );

bool8 boolAttribute(
		const Attribute *		attribute );

bool8 boolBlockAttribute(
		const Block *			block,
		const char *			name );

bool8 * boolCellData(
		const Column *			column,
		unsigned long			rowNumber );

bool8 boolColumnAttribute(
		const Column *			column,
		const char *			name );

bool8 * boolColumnData(
		const Column *			column );

bool8 boolDataSetAttribute(
		const DataSet *			dataSet,
		const char *			name );

bool8 boolTableAttribute(
		const Table *			table,
		const char *			name );

int cellSize(
		Column *			column,
		int				rowNumber );

CellType cellType(
		Column *			column );

bool8 clobber();

const char * columnAttributeComment(
		const Column *			column,
		const char *			name );

AttributeType columnAttributeDataTypeWithName(
		const Column *			column,
		const char *			name );

AttributeType columnAttributeDataTypeWithNumber(
		const Column *			column,
		unsigned int			number );

const char * columnAttributeUnits(
		const Column *			column,
		const char *			name );

Attribute * columnAttributeWithName(
		const Column *			column,
		const char *			name );

Attribute * columnAttributeWithNumber(
		const Column *			column,
		unsigned int			number );

unsigned long columnDataIndex(
		const Column *			column,
		unsigned long			rowNumber,
		unsigned int			numberOfDimensions,
		const unsigned long *		dimensions );

DataType columnDataType(
		const Column *			column );

bool8 columnHasAttribute(
		const Column *			column,
		const char *			name );

unsigned int columnNumber(
		const Table *			table,
		const char *			name );

const char * columnLabel(
		const Column *			column );

Attributable * columnToAttributable(
		Column *			column );

Nullable * columnToNullable(
		Column *			column );

const char * columnUnits(
		const Column *			column );

Column * columnWithName(
		const Table *			table,
		const char *			name );

Column * columnWithNumber(
		const Table *			table,
		unsigned int			position );

const char * commentOfArray(
		Array * 			array,
		unsigned int 			number);

const char * commentOfAttributable(
		Attributable *			attributable,
		unsigned int 			number);

const char * commentOfBlock(
		Block *				block,
		unsigned int 			number);

const char * commentOfColumn(
		Column *			column,
		unsigned int 			number);

const char * commentOfDataSet(
		DataSet *			dataSet,
		unsigned int			number);

const char * commentOfTable(
		Table * 			table,
		unsigned int 			number);

void copyAttributesOfArray(
		Array *				to,
		const Array *			from,
		CopyMode			copyMode );

void copyAttributesOfBlock(
		Block *				to,
		const Block *			from,
		CopyMode			copyMode );

void copyAttributesOfColumn(
		Column *			to,
		const Column *			from,
		CopyMode			copyMode );

void copyAttributesOfDataSet(
		DataSet *			to,
		const DataSet *			from,
		CopyMode			copyMode );

void copyAttributesOfTable(
		Table *				to,
		const Table *			from,
		CopyMode			copyMode );

void copyAttributeToArray(
		Array *				array,
		const Attribute *		attribute );

void copyAttributeToBlock(
		Block *				block,
		const Attribute *		attribute );

void copyAttributeToColumn(
		Column *			column,
		const Attribute *		attribute );

void copyAttributeToDataSet(
		DataSet *			dataSet,
		const Attribute *		attribute );

void copyAttributeToTable(
		Table *				table,
		const Attribute *		attribute );

void copyRows(
		Table *				table,
		unsigned int			from,
		unsigned int			to,
		unsigned long			count );

Nullable * dataComponentToNullable(
		DataComponent *			dataComponent );

void dataServerClient(
		 const char *                   clientName);

DataSet * dataSet(
		const char *			dataSetName,
		AccessMode			openMode,
		MemoryModel			memoryModel );

const char * dataSetAttributeComment(
		const DataSet *			dataSet,
		const char *			name );

AttributeType dataSetAttributeDataTypeWithName(
		const DataSet *			dataSet,
		const char *			name );

AttributeType dataSetAttributeDataTypeWithNumber(
		const DataSet *			dataSet,
		unsigned int			number );

const char * dataSetAttributeUnits(
		const DataSet *			dataSet,
		const char *			name );

Attribute * dataSetAttributeWithName(
		const DataSet *			dataSet,
		const char *			name );

Attribute * dataSetAttributeWithNumber(
		const DataSet *			dataSet,
		unsigned int			number );

bool8 dataSetHasAttribute(
		const DataSet *			dataSet,
		const char *			name );

bool8 dataSetHasBlock(
		const DataSet *			dataSet,
		const char * 			name);

Attributable * dataSetToAttributable(
		DataSet *			dataSet );

void deleteArrayAttributeWithName(
		Array *				array,
		const char *			name );

void deleteArrayAttributeWithNumber(
		Array *				array,
		unsigned int			number );

void deleteAttribute(
		const Attribute *		attribute );

void deleteBlockAttributeWithName(
		Block *				block,
		const char *			name );

void deleteBlockAttributeWithNumber(
		Block *				block,
		unsigned int			number );

void deleteBlockWithName(
		DataSet *			dataSet,
		const char *			name );

void deleteBlockWithNumber(
		DataSet *			dataSet,
		unsigned int			position );

void deleteColumnAttributeWithName(
		Column *			column,
		const char *			name );

void deleteColumnAttributeWithNumber(
		Column *			column,
		unsigned int			number );

void deleteColumnWithName(
		Table *				table,
		const char *			name );

void deleteColumnWithNumber(
		Table *				table,
		unsigned int			position );

void deleteDataSetAttributeWithName(
		DataSet *			dataSet,
		const char *			name );

void deleteDataSetAttributeWithNumber(
		DataSet *			dataSet,
		unsigned int			number );

void deleteRows(
		Table *				table,
		unsigned int			from,
		unsigned int			count );

void deleteTableAttributeWithName(
		Table *				table,
		const char *			name );

void deleteTableAttributeWithNumber(
		Table *				table,
		unsigned int			number );

unsigned long * dimensionsOfArray(
		const Array *			array );

unsigned long * dimensionsOfColumn(
		const Column *			column );

void discardDataSet(
		const char *			dataSetName );

void forEachSubTable(
		const Table *			table,
		TableIteratorFunction 		callThisFunction );

bool8 hasNullsArray(
		const Array *			array );

bool8 hasNullsColumn(
		const Column *			column );

bool8 hasScalingOfArray(
		const Array *			array );

bool8 hasScalingOfColumn(
		const Column *			column );

const char * historyOfArray(
		Array * 			array,
		unsigned int 			number );

const char * historyOfBlock(
		DataSet * 			dataSet,
		unsigned int 			number );

const char * historyOfDataSet(
		DataSet * 			dataSet,
		unsigned int 			number );

const char * historyOfTable(
		Table * 			table,
		unsigned int 			number );

int8 int8ArrayAttribute(
		const Array *			array,
		const char *			name );

int8 * int8ArrayData(
		const Array *			dataSet );

int8 int8Attribute(
		const Attribute *		attribute );

int8 int8BlockAttribute(
		const Block *			block,
		const char *			name );

int8 * int8CellData(
		const Column *			column,
		unsigned long			rowNumber );

int8 int8ColumnAttribute(
		const Column *			column,
		const char *			name );

int8 * int8ColumnData(
		const Column *			column );

int8 int8DataSetAttribute(
		const DataSet *			dataSet,
		const char *			name );

int8 int8TableAttribute(
		const Table *			table,
		const char *			name );

int16 int16ArrayAttribute(
		const Array *			array,
		const char *			name );

int16 * int16ArrayData(
		const Array *			dataSet );

int16 int16Attribute(
		const Attribute *		attribute );

int16 int16BlockAttribute(
		const Block *			block,
		const char *			name );

int16 * int16CellData(
		const Column *			column,
		unsigned long			rowNumber );

int16 int16ColumnAttribute(
		const Column *			column,
		const char *			name );

int16 * int16ColumnData(
		const Column *			column );

int16 int16DataSetAttribute(
		const DataSet *			dataSet,
		const char *			name );

int16 int16TableAttribute(
		const Table *			table,
		const char *			name );

int32 int32ArrayAttribute(
		const Array *			array,
		const char *			name );

int32 * int32ArrayData(
		const Array *			dataSet );

int32 int32AttributableAttribute(
		const Attributable *		attributable,
		const char *			name );

int32 int32Attribute(
		const Attribute *		attribute );

int32 int32BlockAttribute(
		const Block *			block,
		const char *			name );

int32 * int32CellData(
		const Column *			column,
		unsigned long			rowNumber );

int32 int32ColumnAttribute(
		const Column *			column,
		const char *			name );

int32 * int32ColumnData(
		const Column *			column );

int32 int32DataSetAttribute(
		const DataSet *			dataSet,
		const char *			name );

int32 int32TableAttribute(
		const Table *			table,
		const char *			name );

void insertRows(
		Table *				table,
		unsigned int			pos,
		unsigned int			count );

int intNullValueArray(
		const Array *			array );

int intNullValueColumn(
		const Column *			column );

int intNullValueDataComponent(
		const DataComponent *		dataComponent );

int intNullValueNullable(
		const Nullable *		nullable );

bool8 isNullArray(
		const Array *			array,
		unsigned int			pos );

bool8 isNullCell(
		const Column *			column,
		unsigned int			row,
		unsigned int			pos );

bool8 isNullColumn(
		const Column *			column,
		unsigned int			row );

void keepDataSet(
		const char *			dataSetName );

AccessMode mode(
		const DataSet *			dataSet );

const char * nameOfArray(
		const Array * 			array );

const char * nameOfAttributable(
		const Attributable *		attributable );

const char * nameOfAttribute(
		const Attribute *		attribute );

const char * nameOfBlock(
		const Block *			block );

const char * nameOfColumn(
		const Column * 			column );

const char * nameOfDataSet(
		const DataSet * 		dataSet );

const char * nameOfTable(
		const Table * 			table );

bool8 nullDefinedArray(
		const Array *			array );

bool8 nullDefinedColumn(
		const Column *			column );

bool8 nullDefinedDataComponent(
		const DataComponent *		dataComponent );

bool8 nullDefinedNullable(
		const Nullable *		nullable );

NullType nullTypeArray(
		const Array *			array );

NullType nullTypeColumn(
		const Column *			column );

NullType nullTypeDataComponent(
		const DataComponent *		dataComponent );

NullType nullTypeNullable(
		const Nullable *		nullable );

unsigned int numberOfAttributesInArray(
		const Array *			array );

unsigned int numberOfAttributesInAttributable(
		const Attributable *		attributable );

unsigned int numberOfAttributesInBlock(
		const Block *			block );

unsigned int numberOfAttributesInColumn(
		const Column *			column );

unsigned int numberOfAttributesInDataSet(
		const DataSet *			dataSet );

unsigned int numberOfAttributesInTable(
		const Table *			table );

unsigned int numberOfBlocks(
		const DataSet *			dataSet );

unsigned int numberOfColumns(
		const Table *			table );

unsigned int  numberOfCommentsOfArray(
		Array * 			array );

unsigned int  numberOfCommentsOfAttributable(
		Attributable * 			attributable );

unsigned int  numberOfCommentsOfBlock(
		Block * 			block );

unsigned int numberOfCommentsOfDataSet( 
		DataSet *			dataSet );

unsigned int  numberOfCommentsOfTable(
		Table * 			table );

unsigned int numberOfDimensionsOfArray(
		const Array *			array );

unsigned int numberOfDimensionsOfColumn(
		const Column *			column );

unsigned long numberOfElementsOfArray(
		const Array *			array );

unsigned long numberOfElementsOfColumn(
		const Column *			column );

unsigned int  numberOfHistorysOfArray(
		Array * 			array );

unsigned int numberOfHistorysOfBlock( 
		DataSet * 			dataSet );
		
unsigned int numberOfHistorysOfDataSet( 
		DataSet * 			dataSet );
	
unsigned int  numberOfHistorysOfTable(
	      Table * 				table );

unsigned long numberOfRowsOfColumn(
		const Column *			column );

unsigned long numberOfRowsOfTable(
		const Table *			table );

real32 real32ArrayAttribute(
		const Array *			array,
		const char *			name );

real32 * real32ArrayData(
		const Array *			dataSet );

real32 real32Attribute(
		const Attribute *		attribute );

real32 real32BlockAttribute(
		const Block *			block,
		const char *			name );

real32 * real32CellData(
		const Column *			column,
		unsigned long			rowNumber );

real32 real32ColumnAttribute(
		const Column *			column,
		const char *			name );

real32 * real32ColumnData(
		const Column *			column );

real32 real32DataSetAttribute(
		const DataSet *			dataSet,
		const char *			name );

real32 real32TableAttribute(
		const Table *			table,
		const char *			name );

real64 real64ArrayAttribute(
		const Array *			array,
		const char *			name );

real64 * real64ArrayData(
		const Array *			dataSet );

real64 real64Attribute(
		const Attribute *		attribute );

real64 real64BlockAttribute(
		const Block *			block,
		const char *			name );

real64 * real64CellData(
		const Column *			column,
		unsigned long			rowNumber );

real64 real64ColumnAttribute(
		const Column *			column,
		const char *			name );

real64 * real64ColumnData(
		const Column *			column );

real64 real64DataSetAttribute(
		const DataSet *			dataSet,
		const char *			name );

real64 real64TableAttribute(
		const Table *			table,
		const char *			name );

void relabelArray(
		Array *				array,
		const char *			newLabel );

void relabelAttribute(
		Attribute *			attribute,
		const char *			newLabel );

void relabelBlock(
		Block *				block,
		const char *			newLabel );

void relabelColumn(
		Column *			column,
		const char *			newLabel );

void relabelDataSet(
		DataSet *			dataSet,
		const char *			newLabel );

void relabelTable(
		Table *				table,
		const char *			newLabel );

void releaseArray(
		Array *				array );
		
void releaseBlock(
		Block *				block );
		
void releaseColumn(
		Column *			column );
		
void releaseDataSet(
		DataSet *			dataSet );
		
void releaseTable(
		Table *				table );

void renameArray(
		Array *				array,
		const char *			newName );

void renameAttribute(
		Attribute *			attribute,
		const char *			newName );

void renameBlock(
		Block *				block,
		const char *			newName );

void renameColumn(
		Column *			column,
		const char *			newname );

void renameDataSet(
		DataSet *			dataSet,
		const char *			newname );

void renameTable(
		Table *				table,
		const char *			newName );

void setBoolArrayAttribute(
		Array *				array,
		const char *			name,
		bool8				value,
		const char *			comment );

void setBoolAttribute(
		Attribute *			attribute,
		bool8				value,
		const char *			comment );

void setBoolBlockAttribute(
		Block *				block,
		const char *			name,
		bool8				value,
		const char *			comment );

void setBoolColumnAttribute(
		Column *			column,
		const char *			name,
		bool8				value,
		const char *			comment );

void setBoolDataSetAttribute(
		DataSet *			dataSet,
		const char *			name,
		bool8				value,
		const char *			comment );

void setBoolTableAttribute(
		Table *				table,
		const char *			name,
		bool8				value,
		const char *			comment );

void setCellSize(
		Column *			column,
		int				rowNumber,
		int				size );

bool8 setExists(
		const char *			setName );

void setInt8ArrayAttribute(
		Array *				array,
		const char *			name,
		int8				value,
		const char *			units,
		const char *			comment );

void setInt8Attribute(
		Attribute *			attribute,
		int8				value,
		const char *			units,
		const char *			comment );

void setInt8BlockAttribute(
		Block *				block,
		const char *			name,
		int8				value,
		const char *			units,
		const char *			comment );

void setInt8ColumnAttribute(
		Column *			column,
		const char *			name,
		int8				value,
		const char *			units,
		const char *			comment );

void setInt8DataSetAttribute(
		DataSet *			dataSet,
		const char *			name,
		int8				value,
		const char *			units,
		const char *			comment );

void setInt8TableAttribute(
		Table *				table,
		const char *			name,
		int8				value,
		const char *			units,
		const char *			comment );

void setInt16ArrayAttribute(
		Array *				array,
		const char *			name,
		int16				value,
		const char *			units,
		const char *			comment );

void setInt16Attribute(
		Attribute *			attribute,
		int16				value,
		const char *			units,
		const char *			comment );

void setInt16BlockAttribute(
		Block *				block,
		const char *			name,
		int16				value,
		const char *			units,
		const char *			comment );

void setInt16ColumnAttribute(
		Column *			column,
		const char *			name,
		int16				value,
		const char *			units,
		const char *			comment );

void setInt16DataSetAttribute(
		DataSet *			dataSet,
		const char *			name,
		int16				value,
		const char *			units,
		const char *			comment );

void setInt16TableAttribute(
		Table *				table,
		const char *			name,
		int16				value,
		const char *			units,
		const char *			comment );

void setInt32ArrayAttribute(
		Array *				array,
		const char *			name,
		int32				value,
		const char *			units,
		const char *			comment );

void setInt32AttributableAttribute(
		Attributable *			attributable,
		const char *			name,
		int32				value,
		const char *			units,
		const char *			comment );

void setInt32Attribute(
		Attribute *			attribute,
		int32				value,
		const char *			units,
		const char *			comment );

void setInt32BlockAttribute(
		Block *				block,
		const char *			name,
		int32				value,
		const char *			units,
		const char *			comment );

void setInt32ColumnAttribute(
		Column *			column,
		const char *			name,
		int32				value,
		const char *			units,
		const char *			comment );

void setInt32DataSetAttribute(
		DataSet *			dataSet,
		const char *			name,
		int32				value,
		const char *			units,
		const char *			comment );

void setInt32TableAttribute(
		Table *				table,
		const char *			name,
		int32				value,
		const char *			units,
		const char *			comment );

void setNullValueArray(
		Array *				array,
		int				value );

void setNullValueColumn(
		Column *			column,
		int				value );

void setNullValueDataComponent(
		DataComponent *			dataComponent,
		int				value );

void setNullValueNullable(
		Nullable *			nullable,
		int				value );

void setReal32ArrayAttribute(
		Array *				array,
		const char *			name,
		real32				value,
		const char *			units,
		const char *			comment );

void setReal32Attribute(
		Attribute *			attribute,
		real32				value,
		const char *			units,
		const char *			comment );

void setReal32BlockAttribute(
		Block *				block,
		const char *			name,
		real32				value,
		const char *			units,
		const char *			comment );

void setReal32ColumnAttribute(
		Column *			column,
		const char *			name,
		real32				value,
		const char *			units,
		const char *			comment );

void setReal32DataSetAttribute(
		DataSet *			dataSet,
		const char *			name,
		real32				value,
		const char *			units,
		const char *			comment );

void setReal32TableAttribute(
		Table *				table,
		const char *			name,
		real32				value,
		const char *			units,
		const char *			comment );

void setReal64ArrayAttribute(
		Array *				array,
		const char *			name,
		real64				value,
		const char *			units,
		const char *			comment );

void setReal64Attribute(
		Attribute *			attribute,
		real64				value,
		const char *			units,
		const char *			comment );

void setReal64BlockAttribute(
		Block *				block,
		const char *			name,
		real64				value,
		const char *			units,
		const char *			comment );

void setReal64ColumnAttribute(
		Column *			column,
		const char *			name,
		real64				value,
		const char *			units,
		const char *			comment );

void setReal64DataSetAttribute(
		DataSet *			dataSet,
		const char *			name,
		real64				value,
		const char *			units,
		const char *			comment );

void setReal64TableAttribute(
		Table *				table,
		const char *			name,
		real64				value,
		const char *			units,
		const char *			comment );

void setStringArrayAttribute(
		Array *				array,
		const char *			name,
		char *				value,
		const char *			comment );

void setStringAttribute(
		Attribute *			attribute,
		const char *			value,
		const char *			comment );

void setStringBlockAttribute(
		Block *				block,
		const char *			name,
		char *				value,
		const char *			comment );

void setStringColumnAttribute(
		Column *			column,
		const char *			name,
		char *				value,
		const char *			units,
		const char *			comment );

void setStringDataSetAttribute(
		DataSet *			dataSet,
		const char *			name,
		char *				value,
		const char *			comment );

void setStringTableAttribute(
		Table *				table,
		const char *			name,
		char *				value,
		const char *			comment );

void setToNullArray(
		Array *				array,
		unsigned int			pos );

void setToNullCell(
		Column *			column,
		unsigned int			row,
		unsigned int			pos );

void setToNullColumn(
		Column *			column,
		unsigned int			row );

const char * stringArrayAttribute(
		const Array *			array,
		const char *			name );

dstring * stringArrayData(
		const Array *			array );

const char * stringAttribute(
		const Attribute *		attribute );

const char * stringBlockAttribute(
		const Block *			block,
		const char *			name );

const char * stringCellData(
		const Column *			column,
		unsigned long			rowNumber );

const char * stringColumnAttribute(
		const Column *			column,
		const char *			name );

dstring * stringColumnData(
		const Column *			column );

const char * stringDataSetAttribute(
		const DataSet *			dataSet,
		const char *			name );

const char * stringTableAttribute(
		const Table *			table,
		const char *			name );

const char * tableAttributeComment(
		const Table *			table,
		const char *			name );

AttributeType tableAttributeDataTypeWithName(
		const Table *			table,
		const char *			name );

AttributeType tableAttributeDataTypeWithNumber(
		const Table *			table,
		unsigned int			number );

const char * tableAttributeUnits(
		const Table *			table,
		const char *			name );

Attribute * tableAttributeWithName(
		const Table *			table,
		const char *			name );

Attribute * tableAttributeWithNumber(
		const Table *			table,
		unsigned int			number );

bool8 tableHasAttribute(
		const Table *			table,
		const char *			name );

bool8
tableHasColumn(
	       const Table * table,
	       const char * name);

const char * tableLabel(
		const Table *			table );

Attributable * tableToAttributable(
		Table *				table );

Block * tableToBlock(
		Table *				table );

Table * tableWithName(
		const DataSet *			dataSet,
		const char *			name );

Table * tableWithNumber(
		const DataSet *			dataSet,
		unsigned int			position );

uint16 * uint16ArrayData(
		const Array *			array );

uint16 * uint16CellData(
		const Column *			column,
		unsigned long			rowNumber );

uint16 * uint16ColumnData(
		const Column *			column );

uint32 * uint32ArrayData(
		const Array *			array );

uint32 * uint32CellData(
		const Column *			column,
		unsigned long			rowNumber );

uint32 * uint32ColumnData(
		const Column *			column );

/* }; */

/*
forEachBlock
forEachColumn
forEachSubTable
setStringCell
subTable
table
*/


#endif
