// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*************************************************************** 
* 
* DalData.h          & XMM Science Analysis System (SAS) 
* 
* Class representing Data
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: DalData.h,v 1.3 2009/06/22 15:58:03 rsaxton Exp $
* 
*****************************************************************/ 
#ifndef DAL_DATA_H
#define DAL_DATA_H


#include <errstr.h>
#include <Tracer.h>
#include <NAN.h>


#include "Dal.h"
#include "DalChild.h"
#include "DalSeekable.h"
#include "DalDal.h"


template<class T>
class DalData : public virtual Data, public DalDal, public DalChild<T>
{
	public :

		DalData(
			const T *	parent,
			unsigned long	from,
			unsigned long	count,
			AccessMode	accessMode );
		virtual ~DalData() {}

		void setToNull( unsigned long pos ) const;
		bool isNull( unsigned long pos ) const;
		bool hasNulls() const;

		DalSeekable<unsigned long> & seekable() const { PROC RETURN const_cast< DalSeekable<unsigned long> & >( _seekable ); }

		virtual void release() {}

	private :
		DalSeekable<unsigned long>	_seekable;

};


template<class T>
inline DalData<T>::DalData(
	const T *	parent,
	unsigned long	from,
	unsigned long	count,
	AccessMode	accessMode )
	:
		DalDal( accessMode, parent ),
		DalChild<T>( parent ),
		_seekable( from, count )
{
	PROC
	ENDPROC
}

/*AI gcc 4 porting 
parent()->intNullValue();
this->parent()->intNullValue(); OR
DalData<T>::parent() -> intNullValue();
*/
template<class T>
inline void
DalData<T>::setToNull( unsigned long pos ) const
{
	PROC
	switch( dataType() )
	{
		case Data::Bool :

			errstr << "Boolean Null Values are not supported." << Sas::error( "notSupported" );

			break;
		case Data::Int8 :
		{
			int8 * i8 = int8Data();
			i8[ pos ] = this->parent()->intNullValue();
		}
			break;
		case Data::Int16 :
		{
			int16 * i16 = int16Data();
			i16[ pos ] = DalData<T>::parent() -> intNullValue();
		}
			break;
		case Data::Int32 :
		{
			int32 * i32 = int32Data();
			i32[ pos ] = DalData<T>::parent() -> intNullValue();
		}
			break;
		case Data::Real32 :
		{
			real32 * r32 = real32Data();
			NaN.set( r32[ pos ] );
			//r32[ pos ] = parent() -> realNullValue();
		}
			break;
		case Data::Real64 :
		{
			real64 * r64 = real64Data();
			NaN.set( r64[ pos ] );
			//r64[ pos ] = parent() -> realNullValue();
		}
			break;
		case Data::DString :

			errstr << "String Null Values are not supported." << Sas::error( "notSupported" );

			break;
		default :
			errstr << "Invalid data type." << Sas::error( "internalError" );
			break;
	} // switch
	ENDPROC
}


template<class T>
inline bool
DalData<T>::isNull( unsigned long pos ) const
{
	PROC
	bool	isnull = false;

	switch( dataType() )
	{
		case Data::Bool :

			errstr << "Boolean Null Values are not supported." << Sas::error( "notSupported" );

			break;
		case Data::Int8 :
		{
			int8 * i8 = int8Data();
			if( i8[ pos ] == DalData<T>::parent() -> intNullValue() ) isnull = true;
		}
			break;
		case Data::Int16 :
		{
			int16 * i16 = int16Data();
			if( i16[ pos ] == DalData<T>::parent() -> intNullValue() ) isnull = true;
		}
			break;
		case Data::Int32 :
		{
			int32 * i32 = int32Data();
			if( i32[ pos ] == DalData<T>::parent() -> intNullValue() ) isnull = true;
		}
			break;
		case Data::Real32 :
		{
			real32 * r32 = real32Data();
			isnull = NaN.isSet( r32[ pos ] );
			//if( r32[ pos ] == parent() -> realNullValue() ) isnull = true;
		}
			break;
		case Data::Real64 :
		{
			real64 * r64 = real64Data();
			isnull = NaN.isSet( r64[ pos ] );
			//if( r64[ pos ] == parent() -> realNullValue() ) isnull = true;
		}
			break;
		case Data::DString :

			errstr << "String Null Values are not supported." << Sas::error( "notSupported" );

			break;
		default :
			errstr << "Invalid data type." << Sas::error( "internalError" );
			break;
	} // switch
	RETURN isnull;
}

/*AI gcc 4 porting 
if(  hasnulls = isNull( i )  ) break; => if( ( hasnulls = isNull( i ) ) ) break;
suggest parentheses around assignment used as truth value

*/
template<class T>
inline bool
DalData<T>::hasNulls() const
{
	PROC

	bool	hasnulls = false;

	for( unsigned long i = 0; i < _seekable.count(); ++i )
	{
	  if( ( hasnulls = isNull( i ) ) ) break;
	} // for

	RETURN hasnulls;

}

#endif




