// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
using namespace std;
/*************************************************************** 
* 
* dal_utilities.cc           & XMM Science Analysis System (SAS) 
* 
* Various utility functions.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: dal_utilities.cc,v 1.5 2014/04/15 14:44:09 aibarra Exp $ 
* 
*****************************************************************/ 

#include <fitsio.h>
#include <string>
#include <map>
#include <string.h>

#include "dal_utilities.h"
#include "Dal.h"
#include <errstr.h>
#include <Tracer.h>
#include <stringutil.h>
#include "CompoundName.h"

#include <cassert>

// *************************
// Statics are defined here.
// *************************
int numberOfOpenedFitsFiles = 0;


// Global variable used for FITS status.
int status = 0;

/*AI gcc 4 porting 
casting to (char*)
*/
static char * ReservedKeywords1[] =
{
  (char*) "XTENSION", (char*) "BITPIX", (char*)	"NAXIS", (char*) "NAXIS1",
  (char*) "NAXIS2", (char*) "NAXIS3", (char*) "PCOUNT", (char*)	"GCOUNT",
  (char*) "TFIELDS", (char*) "END", (char*) "THEAP", (char*) "EXTNAME",
 (char*) "COMMENT", (char*) "HISTORY", (char*) "BSCALE", (char*) "BZERO",
// (char*) "SIMPLE", (char*) "EXTEND", (char*) "BUNIT", (char*)	"LONGSTRN",
  (char*) "SIMPLE", (char*) "EXTEND", (char*) "BUNIT",
  (char*) "CONTINUE", (char*) "XMMSEP", (char*)	"DATE", (char*)	"CREATOR"
};

/*AI gcc 4 porting 
casting to (char*)
*/
static char * ReservedKeywords2[] =
{
 (char*) "TFORM", (char*) "TTYPE", (char*) "TBCOL", (char*) "TUNIT",
 (char*) "TSCAL", (char*) "TZERO", (char*) "TNULL",
 (char*) "TDIM",  (char*) "XPROC", (char*) "XDAL"

};


static std::string ColumnKeywords[] =
{
	"TLMAX", "TLMIN", "TDMAX", "TDMIN", "TCRVL",
	"TCRPX", "TCDLT", "TCTYP", "TCUNI", "TDISP",
	"TCROT"
};

/*AI gcc 4 porting 
casting to (char*)
*/
static char * AllowedKeywords1[] =
{
  (char*) "HDUCLASS", (char*) "DSTYP",(char*) "DSUNT"
};

static const string nameOfPrimary = "PRIMARY";


//static char * AllowedKeywords2[] =
//{
//	"DSVAL",	"DSREF"
//};


int
toFITSDataType( Dal::DataType	dataType )
{

	PROC

	int	type = TINT;

	switch( dataType )
	{
		case Dal::Bool:

			type = TLOGICAL;

			break;

		case Dal::Int8 :

			type = TBYTE;

			break;

		case Dal::Uint16 :

			type = TUSHORT;

			break;

		case Dal::Int16 :

			type = TSHORT;

			break;

		case Dal::Uint32 :

			type = TUINT;

			break;

		case Dal::Int32 :

			type = TINT;

			break;

		case Dal::Real32 :

			type = TFLOAT;

			break;

		case Dal::Real64 :

			type = TDOUBLE;

			break;

		case Dal::DString :

			type = TSTRING;

			break;

		default :

			errstr << "Encountered an invalid data type " << int( dataType ) << " when trying to convert to a FITS data type" << Sas::error("invalidDataType");

			break;

	} // switch

	RETURN type;

}

/*AI gcc 4 porting 
struct { ... }
*/
struct BB
{
	int		_typecode;
	Dal::DataType	_datatype;
} typecodeMap[] = {
	{ TDOUBLE, 	Dal::Real64 },
	{ TFLOAT, 	Dal::Real32 },
	{ TUSHORT, 	Dal::Uint16 },
	{ TSHORT, 	Dal::Int16 },
	{ TUINT, 	Dal::Uint32 },
	{ TINT, 	Dal::Int32 },
	{ TULONG, 	Dal::Uint32 },
	{ TLONG, 	Dal::Int32 },
	{ TBYTE, 	Dal::Int8 },
	{ TLOGICAL, 	Dal::Bool },
	{ TSTRING, 	Dal::DString }
};


Dal::DataType
fromFITSDataType( int typeCode )
{
	PROC

	static bool		firstTime = true;
	const unsigned int 	nr = sizeof(typecodeMap) / sizeof(typecodeMap[ 0 ]);
	static map<int, Dal::DataType> 	dataTypeOf;

	if( firstTime )
	{
		for( unsigned int i = 0; i < nr; ++i )
		{
			dataTypeOf[ typecodeMap[ i ]._typecode ] = typecodeMap[ i ]._datatype;
		} // for
		firstTime = false;
	} // if

	static map<int,Dal::DataType>::iterator i;

	if( (i = dataTypeOf.find( typeCode )) == dataTypeOf.end() )
	{
		errstr << "Encountered a FITS data type which is not supported by the DAL. The cfitsio type is number " << typeCode << Sas::error("invalidTypeCode");
	} // if

	RETURN (*i).second;

}


bool
isValidDalDataType( int typeCode )
{
	PROC

	static bool		firstTime = true;
	const unsigned int 	nr = sizeof(typecodeMap) / sizeof(typecodeMap[ 0 ]);
	static map<int, Dal::DataType> 	dataTypeOf;

	if( firstTime )
	{
		for( unsigned int i = 0; i < nr; ++i )
		{
			dataTypeOf[ typecodeMap[ i ]._typecode ] = typecodeMap[ i ]._datatype;
		} // for
		firstTime = false;
	} // if

	bool valid = dataTypeOf.find( typeCode ) != dataTypeOf.end();

	RETURN valid;

}


int
toFITSBitpix( Dal::DataType	dataType )
{
	PROC

	int	type = LONG_IMG;

	switch( dataType )
	{
		case Dal::Int8:
			type = BYTE_IMG;
			break;
		case Dal::Uint16:
			type = USHORT_IMG;
			break;
		case Dal::Int16:
			type = SHORT_IMG;
			break;
		case Dal::Uint32:
			type = ULONG_IMG;
			break;
		case Dal::Int32:
			type = LONG_IMG;
			break;
		case Dal::Real32:
			type = FLOAT_IMG;
			break;
		case Dal::Real64 :
			type = DOUBLE_IMG;
			break;
		default :

			errstr << "Encountered an invalid data type " << int( dataType ) << " when trying to convert to a FITS bitpix value" << Sas::error("invalidDataType");

			break;

	} // switch


	RETURN type;

}


Dal::DataType
fromFITSBitpix( int bitpix )
{

	PROC

	Dal::DataType	type = Dal::Int32;

	switch( bitpix )
	{
		case BYTE_IMG :
			type = Dal::Int8;

			break;

		case SHORT_IMG :
			type = Dal::Int16;

			break;

		case LONG_IMG :
			type = Dal::Int32;

			break;

		case FLOAT_IMG :
			type = Dal::Real32;

			break;

		case DOUBLE_IMG :
			type = Dal::Real64;

			break;

		default :

			errstr << "Encountered an invalid FITS bitpix value " << bitpix << Sas::error("invalidTypeCode");

	} // switch

	RETURN type;

}


bool
isFITSReservedKeyword( const string & keyName )
{
	PROC

	bool		found = false;
	unsigned int	nr;
	const string	k = uppercase( keyName );

	//nr = sizeof( ReservedKeywords1 ) / sizeof( char * );
	nr = sizeof( ReservedKeywords1 ) / sizeof( ReservedKeywords1[0]);

	for( unsigned int i = 0; i < nr; i++ )
	{
		if( k == ReservedKeywords1[ i ] )
		{
			found = true;

			break;

		} // if
	} // for

	if( !found )
	{
		//nr = sizeof( ReservedKeywords2 ) / sizeof( char * );
		nr = sizeof( ReservedKeywords2 ) / sizeof(ReservedKeywords2[0] );

		for( unsigned int i = 0; i < nr; i++ )
		{
			if( strncmp(
				ReservedKeywords2[ i ],
				k.c_str(),
				strlen( ReservedKeywords2[ i ] )) == 0 )
			{
				found = true;

				// ***********************************
				// Now check that the remaining
				// characters of the key name form an
				// integer.
				// ***********************************
				for(
					unsigned int j = strlen( ReservedKeywords2[ i ]);
					j < k.length();
					++j )
				{
					if( !isdigit( k[ j ] ))
					{
						found = false;

						break;
					} // if
				} // for

				break;

			} // if
		} // for
	} // if


	RETURN found;

}


bool
isFITSAllowedDigitEndingKeyword( const string & keyName )
{
	PROC

	bool		found = false;
	unsigned int	nr;
	const string	k = uppercase( keyName );

	//nr = sizeof( AllowedKeywords1 ) / sizeof( char * );
	nr = sizeof( AllowedKeywords1 ) / sizeof(AllowedKeywords1[0] );

	for( unsigned int i = 0; i < nr; i++ )
	{
		if( strncmp(
			AllowedKeywords1[ i ],
			k.c_str(),
			strlen( AllowedKeywords1[ i ] )) == 0 )
		{
			found = true;

			// ***********************************
			// Now check that the remaining
			// characters of the key name form an
			// integer.
			// ***********************************
			for(
				unsigned int j = strlen( AllowedKeywords1[ i ]);
				j < k.length();
				++j )
			{
				if( !isdigit( k[ j ] ))
				{
					found = false;

					break;
				} // if
			} // for

			break;

		} // if
	} // for


	RETURN found;

}


bool
isFITSColumnKeyword( const string & keyName )
{
	PROC

	bool		found = false;
	unsigned int	nr;
	const string	k = uppercase( keyName );

	//nr = sizeof( ColumnKeywords ) / sizeof( char * );
	nr = sizeof( ColumnKeywords ) / sizeof( ColumnKeywords[0] );

	for( unsigned int i = 0; i < nr; i++ )
	{
		if( k == ColumnKeywords[ i ] )
		{
			found = true;

			break;

		} // if
	} // for
	RETURN found;
}


int
fits_write_unquoted_string(
		fitsfile * 	fptr,
		char *		keyname,
		char * 		value,
		char * 		comment,
		int * 		status )
{
	PROC

	char		card[ 81 ];

	if( *status != 0 )
	{

		//return *Status;
		ALTRETURNV( 1 );

	} // if

	const int commentLength = 80 - 8 - 3 - strlen( value );

	sprintf(
		card,
		"%-8.8s= %-s%-.*s",
		keyname,
		value,
		commentLength,
		comment );

	fits_write_record( fptr, card, status );

	fits_check(fptr);

	RETURN 0;

}

void fits_check(fitsfile* fptr)
{
	PROC

	char errmsg[1024];

	if( status )
	{

    		errstr << "FITS error " << status;

		if (fptr)
		{
			// =============================================
			// Write out the file name(if possible);
			// there is at least one instance when it is
			// not possible to determine the file name (when
			// the status is FILE_NOT_OPENED), but there may
			// be other circumstances; a switch block is
			// used to handle this.
			// =============================================
			switch( status )
			{
				case FILE_NOT_OPENED :
					errstr << " while accessing a file whose name cannot be determined; the error is probably the filename itself e.g. Blank or contains some illegal characters.";
					break;
				default :
					errstr << " while accessing file '" << fits_file_name( fptr ) << "'";
					break;
			} // switch
		} // if

		fits_get_errstatus(status,errmsg);
		errstr << ": " << errmsg << endl;

		while (fits_read_errmsg(errmsg))
			errstr << errmsg << endl;

		errstr << Sas::error("FITSIO");
	} // if
	else
		// empty the message stack
		while (fits_read_errmsg(errmsg)) ;

	ENDPROC
}


template<class T>
struct AA{
	string			_name;
	T			_type;
};

AA<Attribute::DataType> attributeTypeMap[] =
{
	{ "Bool", 	Attribute::Bool },
	{ "Int", 	Attribute::Int },
	{ "Real", 	Attribute::Real },
	{ "String", 	Attribute::DString }
};

AA<Dal::AccessMode> accessModeTypeMap[] =
{
	{ "Temp",	Dal::Temp },
	{ "Create",	Dal::Create },
	{ "Modify",	Dal::Modify },
	{ "Read",	Dal::Read }
};

AA<Dal::DataType> dalTypeMap[] =
{
	{ "Int8", 	Dal::Int8 },
	{ "Uint16", 	Dal::Uint16 },
	{ "Int16", 	Dal::Int16 },
	{ "Uint32", 	Dal::Uint32 },
	{ "Int32", 	Dal::Int32 },
	{ "Real32", 	Dal::Real32 },
	{ "Real64", 	Dal::Real64 },
	{ "Bool", 	Dal::Bool },
	{ "String", 	Dal::DString }
};

AA<DataSetServer::MemoryModel> memoryModelTypeMap[] =
{
	{ "High",		DataSetServer::High },
	{ "HighLow",		DataSetServer::HighLow },
	{ "Low",		DataSetServer::Low },
	{ "UseEnvironment",	DataSetServer::UseEnvironment }
};


/*
template<class T, AA<T> * S, int nr >
string
toString( T from )
{
	PROC
	static bool				firstTime = true;
	static map<T,string> 			stringRepOf ;

	if( firstTime )
	{
		for( unsigned int i = 0; i < nr; ++i )
		{
			stringRepOf[ S[ i ]._type ] = S[ i ]._name;
		} // for
		firstTime = false;
	} // if

	RETURN stringRepOf[ from ];
}


template string toString<Attribute::DataType, attributeTypeMap, sizeof( attributeTypeMap )/sizeof(attributeTypeMap[0])>( Attribute::DataType );
template string toString<Dal::AccessMode, accessModeTypeMap, sizeof( accessModeTypeMap )/sizeof(accessModeTypeMap[0])>( Dal::AccessMode );
template string toString<Dal::DataType, dalTypeMap, sizeof(dalTypeMap)/sizeof(dalTypeMap[0])>( Dal::DataType );
*/


#define TOSTRING(T, S ) \
string \
toString( T from ) \
{ \
	PROC \
	static bool				firstTime = true; \
	static map<T,string> 			stringRepOf ; \
	const unsigned int			nr = sizeof( S ) / sizeof( S[ 0 ] ); \
 \
	if( firstTime ) \
	{ \
		for( unsigned int i = 0; i < nr; ++i ) \
		{ \
			stringRepOf[ S[ i ]._type ] = S[ i ]._name; \
		} \
		firstTime = false; \
	} \
 \
	RETURN stringRepOf[ from ]; \
}


//template string toString<Attribute::DataType, attributeTypeMap, sizeof( attributeTypeMap )/sizeof(attributeTypeMap[0])>( Attribute::DataType );
//template string toString<Dal::AccessMode, accessModeTypeMap, sizeof( accessModeTypeMap )/sizeof(accessModeTypeMap[0])>( Dal::AccessMode );
//template string toString<Dal::DataType, dalTypeMap, sizeof(dalTypeMap)/sizeof(dalTypeMap[0])>( Dal::DataType );

TOSTRING( Attribute::DataType, attributeTypeMap );
TOSTRING( Dal::DataType, dalTypeMap );
TOSTRING( Dal::AccessMode, accessModeTypeMap );
TOSTRING( DataSetServer::MemoryModel,  memoryModelTypeMap );

/*
string
toString( Attribute::DataType	dataType )
{
	PROC
	static bool				firstTime = true;
	const unsigned int 			nr = sizeof(attributeTypeMap) / sizeof(attributeTypeMap[ 0 ]);
	static map<Attribute::DataType,string> 	stringRepOf ;

	if( firstTime )
	{
		for( unsigned int i = 0; i < nr; ++i )
		{
			stringRepOf[ attributeTypeMap[ i ]._type ] = attributeTypeMap[ i ]._name;
		} // for
		firstTime = false;
	} // if

	RETURN stringRepOf[ dataType ];
}
*/
/*
string
toString( Dal::AccessMode	accessMode )
{
	PROC
	static bool				firstTime = true;
	const unsigned int 			nr = sizeof(accessModeTypeMap) / sizeof(accessModeTypeMap[ 0 ]);
	static map<Dal::AccessMode,string> 	stringRepOf ;

	if( firstTime )
	{
		for( unsigned int i = 0; i < nr; ++i )
		{
			stringRepOf[ accessModeTypeMap[ i ]._type ] = accessModeTypeMap[ i ]._name;
		} // for
		firstTime = false;
	} // if

	RETURN stringRepOf[ dataType ];
}


*/

Attribute::DataType
toAttributeDataType( const string & stringValue )
{

	PROC

	static bool				firstTime = true;
	const unsigned int 			nr = sizeof(attributeTypeMap) / sizeof(attributeTypeMap[ 0 ]);
	static map<string,Attribute::DataType> 	dataTypeOf;

	if( firstTime )
	{
		for( unsigned int i = 0; i < nr; ++i )
		{
			dataTypeOf[ lowercase( attributeTypeMap[ i ]._name ) ] = attributeTypeMap[ i ]._type;
		} // for
		firstTime = false;
	} // if

	static map<string,Attribute::DataType>::iterator 	i;
	if( (i = dataTypeOf.find( lowercase( stringValue ))) == dataTypeOf.end() )
	{
		errstr << "The attribute data type description " << stringValue << " cannot be converted to a data type" << Sas::error("invalidAttributeTypeDescription");

	} // if
	
	const Attribute::DataType	type = (*i).second;

	RETURN type;
}

/*
string
toString( Dal::DataType	dataType )
{
	PROC
	static bool				firstTime = true;
	const unsigned int 			nr = sizeof(dalTypeMap) / sizeof(dalTypeMap[ 0 ]);
	static map<Dal::DataType,string> 	stringRepOf ;

	if( firstTime )
	{
		for( unsigned int i = 0; i < nr; ++i )
		{
			stringRepOf[ dalTypeMap[ i ]._type ] = dalTypeMap[ i ]._name;
		} // for
		firstTime = false;
	} // if

	RETURN stringRepOf[ dataType ];
}

*/

Dal::DataType
toDalDataType( const string & stringValue )
{

	PROC

	static bool				firstTime = true;
	const unsigned int 			nr = sizeof(dalTypeMap) / sizeof(dalTypeMap[ 0 ]);
	static map<string,Dal::DataType> 	dataTypeOf;

	if( firstTime )
	{
		for( unsigned int i = 0; i < nr; ++i )
		{
			dataTypeOf[ lowercase( dalTypeMap[ i ]._name ) ] = dalTypeMap[ i ]._type;
		} // for
		firstTime = false;
	} // if

	static map<string,Dal::DataType>::iterator 	i;
	if( (i = dataTypeOf.find( lowercase( stringValue ))) == dataTypeOf.end() )
	{
		errstr << "The data type description " << stringValue << " cannot be converted to a data type" << Sas::error("invalidTypeDescription");

	} // if
	
	const Dal::DataType	type = (*i).second;

	RETURN type;
}




void
copyAttributes(
		Attributable &		to,
		const Attributable &	from,
		CopyMode		copyMode )
{
	PROC
	switch( copyMode )
	{
		case EraseAllFirst :
			to = from;
			break;
		case Merge :
			break;
		default :
			errstr << "The attributes where not copied because the copy mode '" << int( copyMode ) << "' is invalid " << Sas::error("invalidCopyMode");
	} // switch
	ENDPROC
}



unsigned int
blockNumberToHdunum( const Block & block )
{
	PROC
	unsigned int 		hdunum;
	const Block *		firstBlock = block.parent() -> block( 0 );
	const unsigned int	blockNumber = block.parent() -> blockNumber( block.name() );

	// ********************************************
	// If the first block of the parent dataset is
	// the primary array , then it is neceassary to
	// increment the block number by 1, otherwise
	// increment by 2.
	// ********************************************
	if(
		firstBlock -> type() == Block::ArrayT
		&&
		firstBlock -> name() == nameOfPrimary )
	{
		hdunum = blockNumber + 1;
	}
	else
	{
		hdunum = blockNumber + 2;
	} // if

	RETURN hdunum;
}


unsigned int
byteOffsetOfColumn( fitsfile * /*fptr*/, const Column * column )
{
	PROC
	const Table * table = column -> parent();
	const unsigned int columnNr = table -> columnNumber( column -> name() );
	unsigned int offset = 0;

	for( unsigned int i = 0; i < columnNr; ++i )
	{
		const Column * col = table -> column( i );
		offset += col -> elements() * col -> dataSize();
	} // for
	RETURN offset;

}
 

unsigned long
numberOfElements( const longVec&    size )
{       
	unsigned long n = 0;

	for( unsigned int i = 0; i < size.size(); ++i )
	{
		if( i == 0 )
		{
			n = size[ i ];
		}
		else
		{
			n *= size[ i ];
		} // if
	} // for

	return n;
}


Attribute::DataType
fitsAttributeDataType( const Attribute * attribute )
{
	PROC

	char			dtype;
	Attribute::DataType	dataType = Attribute::Int;

	fits_get_keytype( const_cast<char*>(attribute -> asString().c_str()), &dtype, &status );

	switch( dtype )
	{
		case 'C' : dataType = Attribute::DString; break;
		case 'L' : dataType = Attribute::Bool; break;
		case 'I' : dataType = Attribute::Int; break;
		case 'F' : dataType = Attribute::Real; break;
		case 'X' : errstr << "The attribute has a complex value and is not suppported by the DAL. " << Sas::error( "unsupportedAttributeDataType" );
		default :
			errstr << "Invalid case statement." << Sas::error( "internalError" );
			break;
	} // swtich
	RETURN dataType;
}


template<class T>
class Copier
{
	public :
		virtual ~Copier() {};
		virtual void copyFromInt8( T* /*from*/ ) {};
		virtual void copyFromUint16( T* /*from*/ ) {};
		virtual void copyFromInt16( T* /*from*/ ) {};
		virtual void copyFromUint32( T* /*from*/ ) {};
		virtual void copyFromInt32( T* /*from*/ ) {};
		virtual void copyFromReal32( T* /*from*/ ) {};
		virtual void copyFromReal64( T* /*from*/ ) {};
		//virtual void copy( Array* from, Array* to );
};

/* GCC 3.1 complaint:
   warning: `long unsigned int numberOfElements(...)' 
   defined but not used
*/
/*static*/
unsigned long
numberOfElements( Column* column )
{
	PROC
	RETURN column -> elements() * column -> rows();
}


/*static*/
unsigned long
numberOfElements( Array* array )
{
	PROC
	RETURN array -> elements();
}

template<class F,class T,class C>
static void copy(
	C*			from,
	F * (Data::*fm)() const,
	C*			to,
	T * (Data::*tm)() const )
{
	PROC
	assert( numberOfElements( from ) == numberOfElements( to ) );
	Data * fad = from -> data();
	Data * tad = to -> data();

	F* fd = (fad ->* fm)();
	T* td = (tad ->* tm)();

	const unsigned long nr = numberOfElements( from );

	for( unsigned long i = 0; i < nr; ++i )
	{
		td[ i ] = T( fd[ i ] );
	} // for
	delete fad;
	delete tad;
	ENDPROC
}


template<class T>
class Int8Copier : public virtual Copier<T>
{
	public :
		Int8Copier( T* to ) : _to( to ) {}
		void copyFromInt8( T* from )
		{
			copy<int8,int8,T>(
				from,
				&ArrayData::int8Data,
				_to,
				&ArrayData::int8Data );
		}
		void copyFromUint16( T* from )
		{
			copy<uint16,int8,T>(
				from,
				&ArrayData::uint16Data,
				_to,
				&ArrayData::int8Data );
		}
		void copyFromInt16( T* from )
		{
			copy<int16,int8,T>(
				from,
				&ArrayData::int16Data,
				_to,
				&ArrayData::int8Data );
		}
		void copyFromUint32( T* from )
		{
			copy<uint32,int8,T>(
				from,
				&ArrayData::uint32Data,
				_to,
				&ArrayData::int8Data );
		}
		void copyFromInt32( T* from )
		{
			copy<int32,int8,T>(
				from,
				&ArrayData::int32Data,
				_to,
				&ArrayData::int8Data );
		}
		void copyFromReal32( T* from )
		{
			copy<real32,int8,T>(
				from,
				&ArrayData::real32Data,
				_to,
				&ArrayData::int8Data );
		}
		void copyFromReal64( T* from )
		{
			copy<real64,int8,T>(
				from,
				&ArrayData::real64Data,
				_to,
				&ArrayData::int8Data );
		}
	private :
		T* _to;
};


template<class T>
class Uint16Copier : public virtual Copier<T>
{
	public :
		Uint16Copier( T* to ) : _to( to ) {}
		void copyFromInt8( T* from )
		{
			copy<int8,uint16,T>(
				from,
				&ArrayData::int8Data,
				_to,
				&ArrayData::uint16Data );
		}
		void copyFromUint16( T* from )
		{
			copy<uint16,uint16,T>(
				from,
				&ArrayData::uint16Data,
				_to,
				&ArrayData::uint16Data );
		}
		void copyFromInt16( T* from )
		{
			copy<int16,uint16,T>(
				from,
				&ArrayData::int16Data,
				_to,
				&ArrayData::uint16Data );
		}
		void copyFromUint32( T* from )
		{
			copy<uint32,uint16,T>(
				from,
				&ArrayData::uint32Data,
				_to,
				&ArrayData::uint16Data );
		}
		void copyFromInt32( T* from )
		{
			copy<int32,uint16,T>(
				from,
				&ArrayData::int32Data,
				_to,
				&ArrayData::uint16Data );
		}
		void copyFromReal32( T* from )
		{
			copy<real32,uint16,T>(
				from,
				&ArrayData::real32Data,
				_to,
				&ArrayData::uint16Data );
		}
		void copyFromReal64( T* from )
		{
			copy<real64,uint16,T>(
				from,
				&ArrayData::real64Data,
				_to,
				&ArrayData::uint16Data );
		}
	private :
		T* _to;
};


template<class T>
class Int16Copier : public virtual Copier<T>
{
	public :
		Int16Copier( T* to ) : _to( to ) {}
		void copyFromInt8( T* from )
		{
			copy<int8,int16,T>(
				from,
				&ArrayData::int8Data,
				_to,
				&ArrayData::int16Data );
		}
		void copyFromUint16( T* from )
		{
			copy<uint16,int16,T>(
				from,
				&ArrayData::uint16Data,
				_to,
				&ArrayData::int16Data );
		}
		void copyFromInt16( T* from )
		{
			copy<int16,int16,T>(
				from,
				&ArrayData::int16Data,
				_to,
				&ArrayData::int16Data );
		}
		void copyFromUint32( T* from )
		{
			copy<uint32,int16,T>(
				from,
				&ArrayData::uint32Data,
				_to,
				&ArrayData::int16Data );
		}
		void copyFromInt32( T* from )
		{
			copy<int32,int16,T>(
				from,
				&ArrayData::int32Data,
				_to,
				&ArrayData::int16Data );
		}
		void copyFromReal32( T* from )
		{
			copy<real32,int16,T>(
				from,
				&ArrayData::real32Data,
				_to,
				&ArrayData::int16Data );
		}
		void copyFromReal64( T* from )
		{
			copy<real64,int16,T>(
				from,
				&ArrayData::real64Data,
				_to,
				&ArrayData::int16Data );
		}
	private :
		T* _to;
};


template<class T>
class Uint32Copier : public virtual Copier<T>
{
	public :
		Uint32Copier( T* to ) : _to( to ) {}
		void copyFromInt8( T* from )
		{
			copy<int8,uint32,T>(
				from,
				&ArrayData::int8Data,
				_to,
				&ArrayData::uint32Data );
		}
		void copyFromUint16( T* from )
		{
			copy<uint16,uint32,T>(
				from,
				&ArrayData::uint16Data,
				_to,
				&ArrayData::uint32Data );
		}
		void copyFromInt16( T* from )
		{
			copy<int16,uint32,T>(
				from,
				&ArrayData::int16Data,
				_to,
				&ArrayData::uint32Data );
		}
		void copyFromUint32( T* from )
		{
			copy<uint32,uint32,T>(
				from,
				&ArrayData::uint32Data,
				_to,
				&ArrayData::uint32Data );
		}
		void copyFromInt32( T* from )
		{
			copy<int32,uint32,T>(
				from,
				&ArrayData::int32Data,
				_to,
				&ArrayData::uint32Data );
		}
		void copyFromReal32( T* from )
		{
			copy<real32,uint32,T>(
				from,
				&ArrayData::real32Data,
				_to,
				&ArrayData::uint32Data );
		}
		void copyFromReal64( T* from )
		{
			copy<real64,uint32,T>(
				from,
				&ArrayData::real64Data,
				_to,
				&ArrayData::uint32Data );
		}
	private :
		T* _to;
};


template<class T>
class Int32Copier : public virtual Copier<T>
{
	public :
		Int32Copier( T* to ) : _to( to ) {}
		void copyFromInt8( T* from )
		{
			copy<int8,int32,T>(
				from,
				&ArrayData::int8Data,
				_to,
				&ArrayData::int32Data );
		}
		void copyFromUint16( T* from )
		{
			copy<uint16,int32,T>(
				from,
				&ArrayData::uint16Data,
				_to,
				&ArrayData::int32Data );
		}
		void copyFromInt16( T* from )
		{
			copy<int16,int32,T>(
				from,
				&ArrayData::int16Data,
				_to,
				&ArrayData::int32Data );
		}
		void copyFromUint32( T* from )
		{
			copy<uint32,int32,T>(
				from,
				&ArrayData::uint32Data,
				_to,
				&ArrayData::int32Data );
		}
		void copyFromInt32( T* from )
		{
			copy<int32,int32,T>(
				from,
				&ArrayData::int32Data,
				_to,
				&ArrayData::int32Data );
		}
		void copyFromReal32( T* from )
		{
			copy<real32,int32,T>(
				from,
				&ArrayData::real32Data,
				_to,
				&ArrayData::int32Data );
		}
		void copyFromReal64( T* from )
		{
			copy<real64,int32,T>(
				from,
				&ArrayData::real64Data,
				_to,
				&ArrayData::int32Data );
		}
	private :
		T* _to;
};


template<class T>
class Real32Copier : public virtual Copier<T>
{
	public :
		Real32Copier( T* to ) : _to( to ) {}
		void copyFromInt8( T* from )
		{
			copy<int8,real32,T>(
				from,
				&ArrayData::int8Data,
				_to,
				&ArrayData::real32Data );
		}
		void copyFromUint16( T* from )
		{
			copy<uint16,real32,T>(
				from,
				&ArrayData::uint16Data,
				_to,
				&ArrayData::real32Data );
		}
		void copyFromInt16( T* from )
		{
			copy<int16,real32,T>(
				from,
				&ArrayData::int16Data,
				_to,
				&ArrayData::real32Data );
		}
		void copyFromUint32( T* from )
		{
			copy<uint32,real32,T>(
				from,
				&ArrayData::uint32Data,
				_to,
				&ArrayData::real32Data );
		}
		void copyFromInt32( T* from )
		{
			copy<int32,real32,T>(
				from,
				&ArrayData::int32Data,
				_to,
				&ArrayData::real32Data );
		}
		void copyFromReal32( T* from )
		{
			copy<real32,real32,T>(
				from,
				&ArrayData::real32Data,
				_to,
				&ArrayData::real32Data );
		}
		void copyFromReal64( T* from )
		{
			copy<real64,real32,T>(
				from,
				&ArrayData::real64Data,
				_to,
				&ArrayData::real32Data );
		}
	private :
		T* _to;
};


template<class T>
class Real64Copier : public virtual Copier<T>
{
	public :
		Real64Copier( T* to ) : _to( to ) {}
		void copyFromInt8( T* from )
		{
			copy<int8,real64,T>(
				from,
				&ArrayData::int8Data,
				_to,
				&ArrayData::real64Data );
		}
		void copyFromUint16( T* from )
		{
			copy<uint16,real64,T>(
				from,
				&ArrayData::uint16Data,
				_to,
				&ArrayData::real64Data );
		}
		void copyFromInt16( T* from )
		{
			copy<int16,real64,T>(
				from,
				&ArrayData::int16Data,
				_to,
				&ArrayData::real64Data );
		}
		void copyFromUint32( T* from )
		{
			copy<uint32,real64,T>(
				from,
				&ArrayData::uint32Data,
				_to,
				&ArrayData::real64Data );
		}
		void copyFromInt32( T* from )
		{
			copy<int32,real64,T>(
				from,
				&ArrayData::int32Data,
				_to,
				&ArrayData::real64Data );
		}
		void copyFromReal32( T* from )
		{
			copy<real32,real64,T>(
				from,
				&ArrayData::real32Data,
				_to,
				&ArrayData::real64Data );
		}
		void copyFromReal64( T* from )
		{
			copy<real64,real64,T>(
				from,
				&ArrayData::real64Data,
				_to,
				&ArrayData::real64Data );
		}
	private :
		T* _to;
};


template<class T>
static void
convert( T* object, Dal::DataType toType, T* tmp )
{
	PROC
	// =====================
	// TODO: copy attributes
	// =====================

	Copier<T> * toCopier = 0;

	// **********
	// Copy data.
	// **********
	switch( toType )
	{
		case Dal::Int8 :
			toCopier = new Int8Copier<T>(tmp);
			break;
		case Dal::Uint16 : 
			toCopier = new Uint16Copier<T>(tmp);
			break;
		case Dal::Int16 : 
			toCopier = new Int16Copier<T>(tmp);
			break;
		case Dal::Uint32 : 
			toCopier = new Uint32Copier<T>(tmp);
			break;
		case Dal::Int32 : 
			toCopier = new Int32Copier<T>(tmp);
			break;
		case Dal::Real32 : 
			toCopier = new Real32Copier<T>(tmp);
			break;
		case Dal::Real64 : 
			toCopier = new Real64Copier<T>(tmp);
			break;
		default :
			errstr << "Invalid type" << Sas::error( "internalError" );
			break;
	} // switch

	switch( object -> dataType() )
	{
		case Dal::Int8 :
			toCopier -> copyFromInt8( object );
			break;
		case Dal::Uint16 : 
			toCopier -> copyFromUint16( object );
			break;
		case Dal::Int16 : 
			toCopier -> copyFromInt16( object );
			break;
		case Dal::Uint32 : 
			toCopier -> copyFromUint32( object );
			break;
		case Dal::Int32 : 
			toCopier -> copyFromInt32( object );
			break;
		case Dal::Real32 : 
			toCopier -> copyFromReal32( object );
			break;
		case Dal::Real64 : 
			toCopier -> copyFromReal64( object );
			break;
		default :
			errstr << "Invalid type" << Sas::error( "internalError" );
			break;
	} // switch
	delete toCopier;
	ENDPROC
}


void
convertType( Array* array, Array::DataType toType )
{
	PROC
	// *****************************
	// Return if nothing to be done.
	// *****************************
	if( array -> dataType() == toType )
	{
		ALTRETURN;
	} // if

	// ******************************************************
	// Add a new array to hold the data with the new type.
	// Later this array will be renamed with to the old name.
	// ******************************************************
	Array* tmp = array -> parent() -> addArray(
				"$TEMP$", 
				toType,
				array -> size(),
				array -> units(),
				array -> label(),
				array -> blockNumber() );

	convert( array, toType, tmp );

	const string oldName = array -> name();
	array -> parent() -> deleteBlock( oldName );
	tmp -> rename( oldName );

	ENDPROC
}


void
convertType( Column* column, Column::DataType toType )
{
	PROC
	// *****************************
	// Return if nothing to be done.
	// *****************************
	if( column -> dataType() == toType )
	{
		ALTRETURN;
	} // if

	// ******************************************************
	// Add a new array to hold the data with the new type.
	// Later this array will be renamed with to the old name.
	// ******************************************************
	Column* tmp = column -> parent() -> addColumn(
				"$TEMP$", 
				toType,
				column -> label(),
				column -> units(),
				column -> size(),
				column -> columnNumber() );

	convert( column, toType, tmp );

	const string oldName = column -> name();
	column -> parent() -> deleteColumn( oldName );
	tmp -> rename( oldName );

	ENDPROC
}


static void
dumpData(
	ColumnData*	cd,
	unsigned long	r )
{
	PROC
	switch( cd -> dataType() )
	{
		case Column::Int8 :
		{
			int8* d = cd -> int8Data();
			cout << d[ r ];
		}
			break;
		case Column::Uint16 :
		{
			uint16* d = cd -> uint16Data();
			cout << d[ r ];
		}
			break;
		case Column::Int16 :
		{
			int16* d = cd -> int16Data();
			cout << d[ r ];
		}
		case Column::Uint32 :
		{
			uint32* d = cd -> uint32Data();
			cout << d[ r ];
		}
			break;
		case Column::Int32 :
		{
			int32* d = cd -> int32Data();
			cout << d[ r ];
		}
			break;
		case Column::Real32 :
		{
			real32* d = cd -> real32Data();
			cout << d[ r ];
		}
			break;
		case Column::Real64 :
		{
			real64* d = cd -> real64Data();
			cout << d[ r ];
		}
			break;
		case Column::Bool :
		{
			real64* d = cd -> real64Data();
			d[ r ] ? cout << "T" : cout << "F";
		}
			break;
		case Column::DString :
		{
			char* d = cd -> stringData();
			cout << "\"";
			const unsigned long e = cd->parent()->elements();
			for(unsigned long j=0;j<e;++j)
			{
				cout << d[ r * e + j ];
			} // for
			cout << "\"";
		}
			break;
		default :
			errstr << "Invalid switch" << Sas::error( "internalError" );
	} // switch
	ENDPROC
}


static void
dumpColumnData(
	const map<string,ColumnData*>&	columnWithName,
	ostream&				os )
{
	PROC

	const unsigned long	count = (*(columnWithName.begin())).second -> count();

	for( unsigned int r = 0; r < count; ++r )
	{
		for(
			map<string,ColumnData*>::const_iterator i=columnWithName.begin();
			i!=columnWithName.end();
			++i )
		{
			ColumnData*	cd = (*i).second;
			dumpData( cd, r );
			os << " ";
		} // for
		os << endl;
	} // for

	ENDPROC
}


void
dumpColumns(
	const Table*		table,
	const vector<string>&	columnNames,
	unsigned long		from,
	unsigned long		count,
	ostream&		os )
{
	PROC

	map<string,ColumnData*>	columnWithName;

	if( columnNames.size() )
	{
		for( unsigned int i = 0; i < columnNames.size(); ++i )
		{
			const Column*	column = table -> column( columnNames[ i ] );
			ColumnData*	cd = column -> data( from, count );
			columnWithName[ column -> name() ] = cd;
		} // if
	}
	else
	{
		for( unsigned int i = 0; i < table -> columns(); ++i )
		{
			const Column*	column = table -> column( i );
			ColumnData*	cd = column -> data( from, count );
			columnWithName[ column -> name() ] = cd;
		} // if
	} // if

	dumpColumnData( columnWithName, os );

	for(
		map<string,ColumnData*>::iterator i=columnWithName.begin();
		i!=columnWithName.end();
		++i )
	{
		delete (*i).second;
		//dataSetServer -> close( (*i).second -> parent() -> parent() -> parent() );
	} // if

	ENDPROC
}


struct DataObject {
	DataSet* 	_dataSet;
	Data*		_data;
	string		_name;
	unsigned long 	_count; };


void
dumpObjects(
	const vector<string>&   	objectNames,
	unsigned long           	count,
	bool				fromBeginning,
	ostream&			os )
{
	PROC


	vector<struct DataObject>	dataObjects;

	for( unsigned int i = 0; i < objectNames.size(); ++i )
	{
		const string	object( objectNames[ i ] );
		DataSet*	set( dataSetServer -> open( object, Dal::Read, DataSetServer::Low ) );
		CompoundName	cn( object, CompoundName::Any, set );
		DataObject	dataObject;

		dataObject._dataSet = set;

		switch( cn.state() )
		{
			case CompoundName::BlockName :
			{
				Block*		block = set -> block( object );

				if( block -> type() == Block::TableT )
				{
					Table*			table = set -> table( object );

					for( unsigned int j = 0; j < table -> columns(); ++j )
					{
						Column* column = table -> column( j );
						const unsigned long	from = fromBeginning ? 0 : table -> rows() - count;
						dataObject._name = column -> qualifiedName();

						if( column -> cellType() == Column::Fixed )
						{
							dataObject._data = column -> data( from, count );
							dataObject._count = count;
							dataObjects.push_back( dataObject );
							dataObject._dataSet = 0;
						}
						else
						{
							for( unsigned long k = 0; k < count; ++k )
							{
								CellData* cd( column -> cellData( k + from ));
								dataObject._data = cd;
								dataObject._count = cd -> size();
								dataObjects.push_back( dataObject );
								dataObject._dataSet = 0;
							} // for
						} // if
					} // for
				}
				else
				{
					Array*			array = set -> array( object );
					const unsigned long	from = fromBeginning ? 0 : array -> elements() - count;
					vector<unsigned long> 	dims( array -> size() );
					vector<unsigned long> 	f;
					vector<unsigned long> 	c;

					if( dims.size() == 1 )
					{
						f.push_back( from );
						c.push_back( count );
					}
					else if( dims.size() == 2 )
					{
						f.push_back( ( from + 1 ) / dims[ 1 ] );
						f.push_back( from % dims[ 1 ] );
						c.push_back( ( count + 1 ) / dims[ 1 ] );
						c.push_back( count % dims[ 1 ] );
					}
					else
					{
						f.push_back( ( from + 1 ) / ( dims[ 1 ] * dims[ 2 ] ));
						f.push_back( ( from + 1 ) / ( dims[ 1 ] * dims[ 2 ] ));
						f.push_back( from % dims[ 2 ] );
						c.push_back( ( count + 1 ) / ( dims[ 0 ] * dims[ 2 ] ));
						c.push_back( ( count + 1 ) / ( dims[ 1 ] * dims[ 2 ] ));
						c.push_back( count % dims[ 0 ] );
					} // if
					dataObject._count = count;
					dataObject._name = array -> qualifiedName();
					dataObject._data = array -> data( f, c ) ;
					dataObjects.push_back( dataObject );
				} // if
				break;
			}
			case CompoundName::ColumnName :
			{
				Table*			table = set -> table( object );
				Column* 		column = table -> column( object );
				const unsigned long	from = fromBeginning ? 0 : table -> rows() - count;

				dataObject._name = column -> qualifiedName();

				if( column -> cellType() == Column::Fixed )
				{
					dataObject._data = column -> data( from, count );
					dataObject._count = count;
					dataObjects.push_back( dataObject );
				}
				else
				{
					for( unsigned long k = 0; k < count; ++count )
					{
						CellData* cd( column -> cellData( k + from ));
						dataObject._data = cd;
						dataObject._count = cd -> size();
						dataObjects.push_back( dataObject );
						dataObject._dataSet = 0;
					} // if
				} // if
				break;
				}
			default :
				errstr << "Object type not supported" << Sas::error( "unsupportedObjectType" );
		} // switch
	} // for

	for( unsigned int i = 0; i < dataObjects.size(); ++i )
	{
		DataObject	dataObject( dataObjects[ i ] );
		Data* 		data = dataObject._data;
		unsigned long	c( dataObject._count );

		os << "NAME:" << dataObject._name << " ";

		switch( data -> dataType() )
		{
			case Data::Int8 :
				for( unsigned j = 0; j < c; ++j )
				{
					os << data -> int8Data()[ j ] << " ";
				}
				break;
			case Data::Uint16 :
				for( unsigned j = 0; j < c; ++j )
				{
					os << data -> uint16Data()[ j ] << " ";
				} // for
				break;
			case Data::Int16 :
				for( unsigned j = 0; j < c; ++j )
				{
					os << data -> int16Data()[ j ] << " ";
				} // for
				break;
			case Data::Uint32 :
				for( unsigned j = 0; j < c; ++j )
				{
					os << data -> uint32Data()[ j ] << " ";
				} // for
				break;
			case Data::Int32 :
				for( unsigned j = 0; j < c; ++j )
				{
					os << data -> int32Data()[ j ] << " ";
				} // for
				break;
			case Data::Real32 :
				for( unsigned j = 0; j < c; ++j )
				{
					os << data -> real32Data()[ j ] << " ";
				} // for
				break;
			case Data::Real64 :
				for( unsigned j = 0; j < c; ++j )
				{
					os << data ->real64Data()[ j ] << " ";
				} // for
				break;
			case Data::DString :
				for( unsigned j = 0; j < c; ++j )
				{
					os << data -> stringData()[ j ];
				} // for
				break;
			case Data::Bool :
				for( unsigned j = 0; j < c; ++j )
				{
					os << data -> boolData()[ j ] << " ";
				} // for
				break;
		} // switch

		os << endl;

		// *********************
		// Clean up data object.
		// *********************
		delete data;

	} // for

	os << endl;

	// *******************
	// Close all datasets.
	// *******************
	for( unsigned int i = 0; i < dataObjects.size(); ++i )
	{
		DataSet* set( dataObjects[ i ]._dataSet );

		if( set )
		{
			dataSetServer -> close( set );
		} // if
	}
	ENDPROC
}



string
dalEnvironment()
{
	PROC
	static const char*		envs[] = { "SAS_MEMORY_MODEL", "SAS_ROWS", "SAS_ZERO_ROWS", "SAS_COLUMN_WISE" };
	static const unsigned int	nr = sizeof( envs ) / sizeof( envs[ 0 ] );
	string				env;

	for( unsigned int i = 0; i < nr; ++i )
	{
		env += string( envs[ i ] ) + "=" + (getenv( envs[ i ] ) ?  getenv( envs[ i ] ) : "" ) + " ";
	} // for
	RETURN env;
}


int
fitsOpenFile( fitsfile ** fptr, const char *filename, int iomode, int *status )
{
	PROC

	numberOfOpenedFitsFiles++;

	errstr << "Attempting to open the fits file with name '" << filename << "'. If successful there will be " << numberOfOpenedFitsFiles << " fits files currently open." << Sas::message( Msg::Core, Msg::Sparse );

	RETURN fits_open_file( fptr, filename, iomode, status );
}


int
fitsCloseFile( fitsfile * fptr, int * status )
{
	PROC

	numberOfOpenedFitsFiles--;

	errstr << "Attempting to close the fits file with name '" << fits_file_name( fptr ) << "'. If successful there will be " << numberOfOpenedFitsFiles << " fits files remaining open." << Sas::message( Msg::Core, Msg::Sparse );

	RETURN fits_close_file( fptr, status );
}


int
fitsCreateFile( fitsfile ** fptr, const char *filename, int *status )
{
	PROC

	numberOfOpenedFitsFiles++;

	errstr << "Attempting to create the fits file with name '" << filename << "'. If successful there will be " << numberOfOpenedFitsFiles << " fits files currently open." << Sas::message( Msg::Core, Msg::Sparse );

	RETURN fits_create_file( fptr, filename, status );
}


template<class T,class F>
void
setData(
	Column* to,
	T* (Data::*tdata)() const,
	const Column* from,
	F* (Data::*fdata)() const )
{
	PROC

	Data*		fromColumnData = from -> data();
	F*		fromData = (fromColumnData ->* fdata)();
	Data*		toColumnData = to -> data();
	T*		toData = (toColumnData ->* tdata)();

	if( !from -> nullDefined() && !to -> nullDefined() )
	{
		for( unsigned int i = 0; i < to -> rows() * to -> elements(); ++i )
		{
			toData[ i ] = fromData[ i ];
			// Check for overflow
			if( F( toData[ i ] ) != fromData[ i ] )
			{
				errstr << "While copying the data from column with qualified name '" << from -> qualifiedName() << "' to column with qualified name '" << to -> qualifiedName() << "' an overflow error occurred with element number " << i << "." << Sas::error( "overflow" );
			} // if
		} // for
	}
	else if( !from -> nullDefined() && to -> nullDefined() )
	{
		const int toNullValue = to -> intNullValue();

		for( unsigned int i = 0; i < to -> rows() * to -> elements(); ++i )
		{
			if( int( fromData[ i ] ) == toNullValue )
			{
				errstr << "Null value found" << Sas::error( "nullValueFound" );
			} // if

			toData[ i ] = fromData[ i ];
			// Check for overflow
			if( F( toData[ i ] ) != fromData[ i ] )
			{
				//errstr << "Overflow detected setting to null" << Sas::warning( "overflowDetected" );
				errstr << "While copying the data from column with qualified name '" << from -> qualifiedName() << "' to column with qualified name '" << to -> qualifiedName() << "' an overflow occurred with element number " << i << ". In this case the element will be set to the null value." << Sas::warning( "overflowDetected" );
				toColumnData -> setToNull( i );
			} // if
		} // for
	}
	else if( from -> nullDefined() && !to -> nullDefined() )
	{
		if( fromColumnData -> hasNulls() )
		{
			errstr << "While copying the data from column with qualified name '" << from -> qualifiedName() << "' to column with qualified name '" << to -> qualifiedName() << "' the source column was found to contain null values, but the null value of the destination column is not yet defined" << Sas::error( "unhandledNull" );
		} // if
		for( unsigned int i = 0; i < to -> rows() * to -> elements(); ++i )
		{
			toData[ i ] = fromData[ i ];
			// Check for overflow
			if( F( toData[ i ] ) != fromData[ i ] )
			{
				errstr << "While copying the data from column with qualified name '" << from -> qualifiedName() << "' to column with qualified name '" << to -> qualifiedName() << "' an overflow error occurred with element number " << i << "." << Sas::error( "overflow" );
			} // if
		} // for
	}
	else
	{
		const int toNullValue = to -> intNullValue();
		const int fromNullValue = from -> intNullValue();

		for( unsigned int i = 0; i < to -> rows() * to -> elements(); ++i )
		{
			if( int( fromData[ i ] ) == fromNullValue )
			{
				toData[ i ] = toNullValue;
			}
			else if( int( fromData[ i ] ) == toNullValue )
			{
				errstr << "While copying the data from column with qualified name '" << from -> qualifiedName() << "' to column with qualified name '" << to -> qualifiedName() << "' the element with number " << i << "(in the source column) was found to be equal to the null value of the destination column." << Sas::error( "unhandledNull" );
			}
			else
			{
				toData[ i ] = fromData[ i ];
				// Check for overflow
				if( F( toData[ i ] ) != fromData[ i ] )
				{
					errstr << "While copying the data from column with qualified name '" << from -> qualifiedName() << "' to column with qualified name '" << to -> qualifiedName() << "' an overflow occurred with element number " << i << ". In this case the element will be set to the null value." << Sas::warning( "overflowDetected" );
					toColumnData -> setToNull( i );
				} // if
			} // if

		} // for
	} // if

	delete fromColumnData;
	delete toColumnData;
	ENDPROC
}





template<class T>
void
setData( Column* to, T*(Data::*tdata)() const, const Column* from )
{
	PROC

	switch( from -> dataType() )
	{
		case Column::Int8 :
			setData( to, tdata, from, &Data::int8Data );
			break;
		case Column::Uint16 :
			setData( to, tdata, from, &Data::uint16Data );
			break;
		case Column::Int16 :
			setData( to, tdata, from, &Data::int16Data );
			break;
		case Column::Uint32 :
			setData( to, tdata, from, &Data::uint32Data );
			break;
		case Column::Int32 :
			setData( to, tdata, from, &Data::int32Data );
			break;
		default :
			errstr << "Unhandled data type" << Sas::error( "internalError" );
	} // switch

	ENDPROC
}


void
setData( Column* to, const Column* from )
{
	PROC
	if( to -> rows() != from -> rows() )
	{
		errstr << Sas::error( "incompatibleNumberOfRows" );
	} // if

	switch( to -> dataType() )
	{
		case Column::Int8 :
			setData( to, &Data::int8Data, from );
			break;
		case Column::Uint16 :
			setData( to, &Data::uint16Data, from );
			break;
		case Column::Int16 :
			setData( to, &Data::int16Data, from );
			break;
		case Column::Uint32 :
			setData( to, &Data::uint32Data, from );
			break;
		case Column::Int32 :
			setData( to, &Data::int32Data, from );
			break;
		default :
			errstr << "Unhandled data type" << Sas::error( "internalError" );
	} // switch

	ENDPROC
}


template<class T>
static void
insertVariableColumn(
	Column*		to,
	unsigned long	toRow,
	const Column*	from,
	unsigned long	fromRow,
	unsigned long	count,
	T* (Data::*data)() const )
{
	PROC

	for( unsigned long i = 0; i < count; ++i )
	{
		CellData*	fromCellData = from -> cellData( fromRow + i );
		const T*	fromData = (fromCellData ->* data)();
		CellData*	toCellData = to -> cellData( toRow + i );
		T*		toData = (toCellData ->* data)();

		if( fromCellData -> size() != toCellData -> size() )
		{
			errstr << "Incompatible cells found in variable length columns:" << Sas::error( "incompatibleTables" );

		} // if
		const unsigned long	elements = fromCellData -> size();

		for( unsigned long j = 0; j < elements; ++j )
		{
			toData[ j ] = fromData[ j ];
		} // if
		delete fromCellData;
		delete toCellData;
	} // for

	ENDPROC
}



static void
insertVariableColumn(
	Column*		to,
	unsigned long	toRow,
	const Column*	from,
	unsigned long	fromRow,
	unsigned long	count )
{
	PROC
	switch( to -> dataType() )
	{
		case Column::Int8 :
			insertVariableColumn( to, toRow, from, fromRow, count, &ColumnData::int8Data );
			break;
		case Column::Uint16 :
			insertVariableColumn( to, toRow, from, fromRow, count, &ColumnData::uint16Data );
			break;
		case Column::Int16 :
			insertVariableColumn( to, toRow, from, fromRow, count, &ColumnData::int16Data );
			break;
		case Column::Uint32 :
			insertVariableColumn( to, toRow, from, fromRow, count, &ColumnData::uint32Data );
			break;
		case Column::Int32 :
			insertVariableColumn( to, toRow, from, fromRow, count, &ColumnData::int32Data );
			break;
		case Column::Real32 :
			insertVariableColumn( to, toRow, from, fromRow, count, &ColumnData::real32Data );
			break;
		case Column::Real64 :
			insertVariableColumn( to, toRow, from, fromRow, count, &ColumnData::real64Data );
			break;
		case Column::Bool :
			insertVariableColumn( to, toRow, from, fromRow, count, &ColumnData::boolData );
			break;
		case Column::DString :
			insertVariableColumn( to, toRow, from, fromRow, count, &ColumnData::stringData );
			break;
		default :
			errstr << "Unhandled data type" << Sas::error( "internalError" );
	} // switch
	ENDPROC
}


template<class T>
static void
insertFixedColumn(
	Column*		to,
	unsigned long	toRow,
	const Column*	from,
	unsigned long	fromRow,
	unsigned long	count,
	T* (Data::*data)() const )
{
	PROC

	const Data*		fromColumnData = from -> data( fromRow, count );
	const T*		fromData = (fromColumnData ->* data)();
	Data*			toColumnData = to -> data( toRow, count );
	T*			toData = (toColumnData ->* data)();
	const unsigned long	elements = count * to -> elements();

	for( unsigned long i = 0; i < elements; ++i )
	{
		toData[ i ] = fromData[ i ];
	} // for

	ENDPROC
}


static void
insertFixedColumn(
	Column*		to,
	unsigned long	toRow,
	const Column*	from,
	unsigned long	fromRow,
	unsigned long	count )
{
	PROC
	switch( to -> dataType() )
	{
		case Column::Int8 :
			insertFixedColumn( to, toRow, from, fromRow, count, &Data::int8Data );
			break;
		case Column::Uint16 :
			insertFixedColumn( to, toRow, from, fromRow, count, &Data::uint16Data );
			break;
		case Column::Int16 :
			insertFixedColumn( to, toRow, from, fromRow, count, &Data::int16Data );
			break;
		case Column::Uint32 :
			insertFixedColumn( to, toRow, from, fromRow, count, &Data::uint32Data );
			break;
		case Column::Int32 :
			insertFixedColumn( to, toRow, from, fromRow, count, &Data::int32Data );
			break;
		case Column::Real32 :
			insertFixedColumn( to, toRow, from, fromRow, count, &Data::real32Data );
			break;
		case Column::Real64 :
			insertFixedColumn( to, toRow, from, fromRow, count, &Data::real64Data );
			break;
		case Column::Bool :
			insertFixedColumn( to, toRow, from, fromRow, count, &Data::boolData );
			break;
		case Column::DString :
			insertFixedColumn( to, toRow, from, fromRow, count, &Data::stringData );
			break;
		default :
			errstr << "Unhandled data type" << Sas::error( "internalError" );
	} // switch
	ENDPROC
}


static void
insertColumn(
	Column*		to,
	unsigned long	toRow,
	const Column*	from,
	unsigned long	fromRow,
	unsigned long	count )
{
	PROC
	if( to -> cellType() == Column::Fixed )
	{
		insertFixedColumn( to, toRow, from, fromRow, count );
	}
	else
	{
		insertVariableColumn( to, toRow, from, fromRow, count );
	} // if
	ENDPROC
}


void
insertTable(
	Table*		to,
	unsigned long	toRow,
	const Table*	from,
	unsigned long	fromRow,
	unsigned long	count )
{
	PROC
	if( to -> columns() != from -> columns() )
	{
		errstr << "The table with qualified name '" << to -> qualifiedName() << "' has " <<  to -> columns() << " columns, but " << from -> columns() << " were expected" << Sas::error( "incompatibleTables" );

	} // if

	to -> insertRows( toRow, count );

	for( unsigned int i = 0; i < from -> columns(); ++i )
	{
		Column*		toColumn = to -> column( i );
		const Column*	fromColumn = from -> column( i );

		if( fromColumn -> name() != toColumn -> name() )
		{
			errstr << "The table with qualified name '" << to -> qualifiedName() << "' contains a column with name '" << toColumn -> name() << "' in position " << i << ", but a column with name '" << fromColumn -> name() << "' was expected." << Sas::error( "incompatibleTables" );
		} // if
		if( toColumn -> dataType() != fromColumn -> dataType() )
		{
			errstr << "The column with qualified name '" << toColumn -> qualifiedName() << "' has datatype " << toString( toColumn -> dataType() ) << " but it was expected to be " << toString( fromColumn -> dataType() ) << "." << Sas::error( "incompatibleTables" );
		} // if
		if( toColumn -> cellType() != fromColumn -> cellType() )
		{
			errstr << "The columns with qualifed names '" << toColumn -> qualifiedName() << "' and '" << fromColumn -> qualifiedName() << "' have different cell types" << Sas::error( "incompatibleTables" );
		} // if
		if( toColumn -> size() != fromColumn -> size() )
		{
			errstr << "The columns with qualifed names '" << toColumn -> qualifiedName() << "' and '" << fromColumn -> qualifiedName() << "' have different cell sizes" << Sas::error( "incompatibleTables" );
		} // if
		insertColumn( toColumn, toRow, fromColumn, fromRow, count );
	} // for

	ENDPROC
}


//unsigned long
//sasRows()
//{
//	PROC
//	const char * 		dal_rows = getenv( "SAS_ROWS" );
//	const unsigned long	rows = (dal_rows ? atoi( dal_rows ) : 1 );

//	RETURN rows;
//}


// ***********************************
// TableSeeker is defined beyond here.
// ***********************************
TableSeeker::TableSeeker( Table* table ) :
		_from( 0 ),
		_count( 0 ),
		_sasRows( 0 ),
		_more( true ),
		_table( table ),
		_ofrom( table -> from() ),
		_ocount( table -> count() )
{
	const char* 	sasRows = getenv( "SAS_ROWS" );
	_sasRows = ( sasRows ? atoi( sasRows ) : 1 );
}


TableSeeker::~TableSeeker()
{
	_table -> seek( _ofrom, _ocount );
}


bool
TableSeeker::next()
{
	if( !_more )
	{
		return false;
	} // if

	const unsigned long 	r = _table -> rows();

	_from += _count;

	if( _from + _sasRows < r )
	{
		_count = _sasRows;
	}
	else
	{
		_count = r - _from;
		_more = false;
	} // if

	_table -> seek( _from, _count );

	return true;
}


/*
bool
TableSeeker::next()
{
	if( !_more )
	{
		return false;
	} // if

	const unsigned long 	r = _table -> rows();

	// If count == 0 then this is the first time
	if( _count == 0 )
	{
		// First time
		if( _from + _sasRows < r )
		{
			_count = _sasRows;
		}
		else
		{
			_count = r;
			_more = false;
		} // if
	}
	else
	{
		// Not first time
		_from += _count;

		if( _from + _sasRows < r )
		{
			_count = _sasRows;
		}
		else
		{
			_count = r - _from;
			_more = false;
		} // if
	} // if

	_table -> seek( _from, _count );

	return true;
}
*/


void
TableSeeker::reset()
{
	_from = 0;
	_count = 0;
	_more = true;
	_table -> seek( _ofrom, _ocount );
}


unsigned long
TableSeeker::from() const
{
	PROC
	RETURN _from;
}


unsigned long
TableSeeker::count() const
{
	PROC
	RETURN _count;
}



