// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*************************************************************** 
* 
* dal_utilites.h             & XMM Science Analysis System (SAS) 
* 
* Various utility functions.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: dal_utilities.h,v 1.2 2004/03/05 14:43:01 eojero Exp $ 
* 
*****************************************************************/ 
#ifndef DAL_UTILITIES_H
#define DAL_UTILITIES_H


#include <fitsio.h>
#include <string>

#include <errstr.h>
#include <Tracer.h>

#include "Dal.h"

class Converter
{
	public :


		Attribute::DataType
		toAttributeDataType( const std::string & stringValue )
		{
			PROC

			Attribute::DataType	type = Attribute::Int; // w/o initialization
													   // optimizer complains

			if( stringValue == "int" )
			{
				type = Attribute::Int;
			}
			else if( stringValue == "real" )
			{
				type = Attribute::Real ;
			}
			else if( stringValue == "bool" )
			{
				type = Attribute::Bool;
			}
			else if( stringValue == "string" )
			{
				type = Attribute::DString;
			}
			else
			{
				errstr << "The data type description " << stringValue << " for an attribute cannot be converted to a data type" << Sas::error("invalidAttributeTypeDescription");

			} // if

			RETURN type;
		}

		int toFITSDataType( Attribute::DataType dataType )
		{
			PROC

			int	type;

			switch( dataType )
			{
				case Attribute::Int :

					type = TLONG; break;

				case Attribute::Real :

					type = TDOUBLE; break;

				case Attribute::DString :

					type = TSTRING; break;

				case Attribute::Bool :

					type = TLOGICAL; break;

				default :

					errstr << "Encountered an invalid data type " << int( dataType ) << " when trying to convert to a FITS data type" << Sas::error("invalidDataType");

					break;

			} // switch

			RETURN type;

		}

};


Data::DataType toDalDataType( const std::string & stringValue );

Attribute::DataType toAttributeDataType( const std::string & stringValue );

std::string toString( Dal::DataType dataType );

std::string toString( Attribute::DataType dataType );

std::string toString( Dal::AccessMode accessMode );

std::string toString( DataSetServer::MemoryModel memoryModel );

int toFITSDataType( Data::DataType 		dataType );

bool isValidDalDataType( int typecode );

Data::DataType fromFITSDataType( int typeCode );

bool isFITSReservedKeyword( const std::string & keyName );

bool isFITSAllowedDigitEndingKeyword( const std::string & keyName );

int toFITSBitpix( Data::DataType		dataType );

bool isFITSColumnKeyword( const std::string & keyName );

Data::DataType fromFITSBitpix( int bitpix );

int fits_write_unquoted_string(
		fitsfile * 	fptr,
		char *		keyname,
		char * 		value,
		char * 		comment,
		int * 		status );


extern int status;

void fits_check(fitsfile *fptr=0);

enum CopyMode { EraseAllFirst = 1, Merge };

void copyAttributes(
		Attributable &	to,
		const Attributable &	from,
		CopyMode		copyMode = EraseAllFirst );


// *****************************************************
// Returns the number of elements in the size decriptor.
// *****************************************************
unsigned long numberOfElements( const longVec & size );


// *****************************************************
// Determine the fits header number, of the given block.
// *****************************************************
unsigned int blockNumberToHdunum( const Block & block );

// ********************************************************
// Returns the byte position of the specified column in the
// given fits file.
// ********************************************************
unsigned int byteOffsetOfColumn( fitsfile * fptr, const Column * column );


// *****************************************************************
// Define the cfitsio v2.0 fits_file_name function in the event that
// an earlier version is being used.
// BAD_DATE is an new symbolic name, introduced in cfitsio v2.0, so
// this may be used to establish if an earlier version is
// used.
// *****************************************************************
#ifndef BAD_DATE
inline std::string fits_file_name( fitsfile * fptr ) { return fptr -> filename; }
#else
inline std::string fits_file_name( fitsfile * fptr ) { char name[ 256 ]; int status; ffflnm( fptr, name, &status ); return name; }
#endif


// **************************************************************
// Determine the data type of the given attribute, using the fits
// convenience function fits_get_keytypei() / ffdtyp().
// **************************************************************
Attribute::DataType fitsAttributeDataType( const Attribute * attribute );


// ***************************************************
// Convert the given array to the specified data type.
// ***************************************************
void convertType( Array* array, Array::DataType toType );


// ****************************************************
// Convert the given column to the specified data type.
// ****************************************************
void convertType( Column* column, Column::DataType toType );


// *************************************************
// Dump an ascii representation of the names columns
// to the specified stream.
// *************************************************
void dumpColumns(
	const Table*            	table,
	const std::vector<std::string>&   	columnNames,
	unsigned long           	from, 
	unsigned long           	count,
	ostream&			os );


// *************************************************
// Dump an ascii representation of the named data
// objects to the specified stream.
// *************************************************
void dumpObjects(
	const std::vector<std::string>&   	objectNames,
	unsigned long           	count,
	bool				fromBeginning,
	ostream&			os );


// *************************************************
// Return a string representing the DAL environment.
// *************************************************
std::string dalEnvironment();
 

// ************************************************
// The following is essentially for debug purposes,
// and is designed to keep track of the number
// of fits files which are currently open.
// ************************************************
//static int numberOfOpenedFitsFiles;
int fitsOpenFile( fitsfile ** fptr, const char *filename, int iomode, int *status );
int fitsCloseFile( fitsfile * fptr, int * status );
int fitsCreateFile( fitsfile ** fptr, const char *filename, int *status );


void setData( Column *	to, const Column *	from );

void insertTable(
	Table*		to,
	unsigned long	toRow,
	const Table*	from,
	unsigned long	fromRow,
	unsigned long	count );


// ************************************************************************
// DESCRIPTION
// ===========
// A utility class that makes it easy to traverse the rows of a table
// in a low memory friendly manner.
// Something like this could alternatively be part of the extended dal
// METHODS
// =======
// bool next()
//	Moves to the next seek range (if there are any more). REturns true
//	if there was another seek range to move to, false otherwise
// void initialise()
//	 Resets the seeker.
// unsigned long from()
//	const returns the from value of the current seek range
// unsigned long count()
//	const returns the count value of the current seek range
// ************************************************************************

class TableSeeker
{
	public :
		TableSeeker( Table* table );
		~TableSeeker();

		bool			next();
		void			reset();
		unsigned long		from() const;
		unsigned long		count() const;

	private :
		unsigned long		_from;
		unsigned long		_count;
		unsigned long		_sasRows;
		bool			_more;
		Table*			_table;
		const unsigned long	_ofrom;
		const unsigned long	_ocount;
};


#endif
