// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
using namespace std;
/*************************************************************** 
* 
* HiLowDataSet.cc              & XMM Science Analysis System (SAS) 
* 
* Class representing a DAL data set.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: HiLowDataSet.cc,v 1.2 2004/03/05 14:43:02 eojero Exp $ 
* 
*****************************************************************/ 

#include <fstream>			// C++

#include <errstr.h>			// SAS
#include <Tracer.h>

#include "Dal.h"			// DAL
#include "HiLowDataSet.h"
#include "HiLowTable.h"
#include "HiLowArray.h"
#include "DataSetReaderWriter.h"


void
HiLowDataSet::readmodify()
{
	PROC

	errstr << "Reading dataset with name '" << name() << "'" << Sas::message( Msg::Core, Msg::Verbose );

	// **************************
	// Lock the Database manager.
	// **************************
	_dataSetReaderWriter -> lock();

	alwaysAllowModify( true );

	// **************
	// Read the file.
	// **************
	_dataSetReaderWriter -> read( this );

	// ****************************
	// Unlock the Database manager.
	// ****************************
	_dataSetReaderWriter -> unlock();

	alwaysAllowModify( false );

	ENDPROC

}


void
HiLowDataSet::createtemp( DataSetServer * server )
{
	PROC

	errstr << "Creating dataset with name '" << name() << "'" << Sas::message( Msg::Core, Msg::Verbose );

	_dataSetReaderWriter -> clobber( server -> clobber() );


	// ****************************
	// Unlock the Database manager.
	// ****************************
	_dataSetReaderWriter -> unlock();

	_dataSetReaderWriter -> addEntry( *this );

	ENDPROC
}


HiLowDataSet::HiLowDataSet(
			const string &			setName,
			AccessMode			openMode,
			DalDataSetServer *		server,
			DataSetReaderWriter*		readerWriter )
	:
		DalDataSet( setName, "", openMode, DataSetServer::HighLow, server )
{

	PROC

	errstr << "Accessing dataset with name '" << name() << "'" << Sas::message( Msg::Core, Msg::Verbose );

	_dataSetReaderWriter = readerWriter;

	addDebug();

	switch( mode() )
	{
		case Read :
		case Modify :

			readmodify();

			break;

		case Temp :
		case Create :

			createtemp( server );

			break;

		//case Temp :

			// ********************************
			// TODO: PROPERLY IMPLEMENTED
			// ********************************
			//DEBUG( clog << "creating temporary data set " << name() << endl; )

			//break;


		default :

			errstr << "The DAL could not create data set '" << name() << "' since its access mode " << int( openMode ) <<" is invalid" << Sas::error("internalError");

			break;


	} // switch


	ENDPROC

}


HiLowDataSet::HiLowDataSet( const HiLowDataSet & dataSet )
	:
  Dal(),
  Labelled(),
  Attributable(),
  Child<DataSetServer>(),
  DataSet(),
  Child<DalDataSetServer>(),
  DalDataSet( name(), label(), mode(), memoryModel(), DalChild<DalDataSetServer>::parent() )
{
	PROC
	_blocks = dataSet._blocks;

	//switch( dataSet._dataSetReaderWriter -> fileType() )
	//{
	//	case DataSetServer::Fits :
	//		_dataSetReaderWriter = new HiLowFitsReaderWriter( this );
	//		break;

	//	case DataSetServer::Dal :
	//		_dataSetReaderWriter = new HiLowDalReaderWriter( this );
	//		break;

	//	default :
	//		errstr << "The file type (with value " << int( dataSet._dataSetReaderWriter -> fileType() ) << ") is not recognised" << Sas::error( "InternalError" );
	//		break;

	//} // switch

	ENDPROC
}


//DalDataSet&
//HiLowDataSet::operator=( const DalDataSet& /*dataSet*/ )
//{
//	PROC
	//(Attributable &)*this = dataSet;
	//(DalDataSet &)*this = dataSet;
	// TODO: IMPLEMENT
	//_blocks = dataSet.blocks();

	//switch( dataSet.fileReaderWriter() -> fileType() )
	//{
	//	case DataSetServer::Fits :

	//		_dataSetReaderWriter = new HiLowFitsFileReaderWriter( this );

	//		break;

	//	case DataSetServer::Dal :

	//		_dataSetReaderWriter = new HiLowDalFileReaderWriter( this );

	//		break;

	//	default :

	//		errstr << "The file type (with value " << int( dataSet.fileReaderWriter() -> fileType() ) << ") is not recognised" << Sas::error( "InternalError" );

	//		break;

	//} // switch

//	RETURN *this;
//}


HiLowDataSet::~HiLowDataSet()
{
	PROC

	switch( mode() )
	{
		case Create :
		case Modify :


			// ***************
			// Write the file.
			// ***************
			_dataSetReaderWriter -> write( this );

			break;


		case Read :

			break;

		case Temp :

			// ****************
			// Delete the file.
			// ****************
			_dataSetReaderWriter -> deleteEntry( *this );

			break;

		default :

			errstr << "The DAL could not close data set '" << name() << "' since its access mode " << int( mode() ) <<" is invalid" << Sas::error("internalError");

			break;

	} // switch

	// ******************************
	// Release and delete the blocks.
	// ******************************
	release();

	for(
		BlockList::iterator i = _blocks.begin();
		i != _blocks.end();
		++i )
	{
		delete (*i);
	} // for

	delete _dataSetReaderWriter;

	ENDPROC

}


DalTable *
HiLowDataSet::makeTable(
			const string &		name,
			const string &		label,
			unsigned long		rows,
			DalDataSet *		parent )
{
	PROC
	RETURN new HiLowTable( name, rows, label, parent );
}


DalArray *
HiLowDataSet::makeArray(
			const string &			name,
			DataType			dataType,
			const longVec&			size,
			const string &			units,
			const string &			label,
			DalDataSet *			parent )
{
	PROC
	RETURN new HiLowArray( name, dataType, size, units, label, parent );
}


void
HiLowDataSet::printOn( ostream & os ) const
{
	PROC

	// ************************
	// Write the Data Set Name.
	// ************************
	os << "DATA SET: Name->" << name() << endl;

	// **************************
	// Write the Data Set Header.
	// **************************
	DalAttributable::printOn( os );


	// **************************
	// Write the Data Set Blocks.
	// **************************
	os << "There are " << _blocks.size() << " blocks" << endl;

	unsigned int no = 1;

	for(
		BlockList::const_iterator i = _blocks.begin();
		i != _blocks.end();
		++i, ++no )
	{
		os << "[" << no << "]";

		(*i) -> printOn( os );

	} // for

	ENDPROC
}


