// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
using namespace std;
/*************************************************************** 
* 
* DataStorage.cc             & XMM Science Analysis System (SAS) 
* 
* Class representing data storage object.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: DataStorage.cc,v 1.2 2004/03/05 14:43:02 eojero Exp $ 
* 
*****************************************************************/ 


#include "DataStorage.h"

#include <errstr.h>
#include <Tracer.h>

#include <string>
#include <vector>

#include "Dal.h"
#include "DalChild.h"
#include "DalSeekable.h"
#include "DalDal.h"
#include "dal_utilities.h"


class Values : public DalDal, public DalChild<DataStorage>, public DalSeekable<unsigned long>
{
	public :

		// Constructors
		Values( 
			unsigned long			size,
			DataStorage*			parent );

		virtual ~Values() {}

		virtual int8 * int8Values( unsigned long, unsigned long )
			 { PROC invalidDataType(Int8); RETURN static_cast<int8*>(0); }

		virtual uint16 * uint16Values( unsigned long, unsigned long )
			 { PROC invalidDataType(Uint16); RETURN static_cast<uint16*>(0); }

		virtual int16 * int16Values( unsigned long, unsigned long )
			 { PROC invalidDataType(Int16); RETURN static_cast<int16*>(0); }

		virtual uint32 * uint32Values( unsigned long, unsigned long )
			 { PROC invalidDataType(Uint32); RETURN static_cast<uint32*>(0); }

		virtual int32 * int32Values( unsigned long, unsigned long )
			 { PROC invalidDataType(Int32); RETURN static_cast<int32*>(0); }

		virtual real32 * real32Values( unsigned long, unsigned long )
			 { PROC invalidDataType(Real32); RETURN static_cast<real32*>(0); }

		virtual real64 * real64Values( unsigned long, unsigned long )
			 { PROC invalidDataType(Real64); RETURN static_cast<real64*>(0); }

		virtual bool8 * boolValues( unsigned long, unsigned long )
			 { PROC invalidDataType(Bool); RETURN static_cast<bool8*>(0); }

		virtual char * stringValues( unsigned long, unsigned long )
			 { PROC invalidDataType(DString); RETURN static_cast<char*>(0); }

		virtual unsigned int dataSize() const = 0;
		virtual void size( unsigned long size ) = 0;
		unsigned long size() const { PROC RETURN _size; }
		virtual DataType dataType() const = 0;
		virtual void printOn( ostream& os, const string & sep="" ) const = 0;
		virtual void insertBytes( unsigned long pos, unsigned long count=1 ) { PROC pos=count; errstr << int( dataType() ) << Sas::error("notImplemented"); ENDPROC }
		virtual void deleteBytes( unsigned long from, unsigned long count=1 ) { PROC from=count; Sas::error("notImplemented",""); ENDPROC }
		virtual void copyBytes( unsigned long from, unsigned long to, unsigned long count=1 ) { PROC from=to=count; Sas::error("notImplemented",""); ENDPROC } 
		void release() { PROC Sas::error("notImplemented",""); ENDPROC } 

	protected :
		unsigned long		_size;
	private:

		void invalidDataType( Dal::DataType requestedType )
		{
			PROC
			errstr << "It is not possible to return the requested data (requested type = " << int( requestedType ) << "), as it is of the wrong type (actual type = " << int( dataType() ) << ")." << Sas::error( "invalidDataType" );
			ENDPROC
		}

};


template<class T>
void writeValues(
		ostream&	os,
		T *		values,
		unsigned long	numberOfValues,
		const string &	sep="" )
{
	PROC

	for( unsigned long j = 0; j < numberOfValues; ++j )
	{
		os << values[ j ] << sep;
	} // for

	os << endl;

	ENDPROC
}


template<class T>
void
insertBytes( T* &values, unsigned long& size, unsigned long pos, unsigned long count )
{
	PROC

	if( pos >= size ) pos = size;

	T* newValues = new T[ size + count ];

	for( unsigned long i = 0; i < pos; ++i )
	{
		newValues[ i ] = values[ i ];
	} // for

	for( unsigned long i = pos; i < size; ++i )
	{
		newValues[ count + i ] = values[ i ];
	} // for

	delete[] values;

	values = newValues;

	size = size + count;
}


template<class T>
void deleteBytes( T* &values, unsigned long& size, unsigned long from, unsigned long count )
{
	PROC

	if ( from >= size ) ALTRETURN;

	count = min( size - from, count );

	T* newValues = new T[ size - count ];

	for( unsigned long i = 0; i < from; ++i )
	{
		newValues[ i ] = values[ i ];
	} // for

	for( unsigned long i = from + count; i < size; ++i )
	{
		newValues[ i - count ] = values[ i ];
	} // for

	delete[] values;

	values = newValues;

	size = size - count;

}


template<class T>
void copyBytes( T* values, unsigned long size, unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	if ( from >= size || to >= size )
	{
        	SAS_DEBUG( clog << "copyRows cancelled: out of range" << endl; )
        	ALTRETURN
        } // if

	count = min( count, min( size - from, size - to ) );

	// Handle overlapping regions by copying in the right direction.
	unsigned long	nFrom;
	unsigned long	nTo;

	// If the starting destination is inside the source range,
        // copy starting at the last element.
	if ( to >= from && to < from + count )
	{

		// Copy starting at the last element.
		nFrom = from + count - 1;
		nTo = to + count - 1;

		for(
			unsigned long i = 0, j = nFrom, k = nTo;
			i < count;
			++i, --j, --k )
		{
			values[k] = values[j];
		} // for
	}
	else
	{

		// Copy starting at the first element
		nFrom = from;
		nTo = to;

		for(
			unsigned long i = 0, j = nFrom, k = nTo;
			i < count;
			++i, ++j, ++k )
		{
			values[k] = values[j];
		} // for
	} // if

	ENDPROC
}


class BoolValues : public virtual Values
{
	public :

		// Constructors
		BoolValues(
			unsigned long			size,
			DataStorage *			parent );

		virtual ~BoolValues();
		bool8 * boolValues( unsigned long from, unsigned long count );
		unsigned int dataSize() const;
		DataType dataType() const;
		void size( unsigned long size );
		void printOn( ostream& os, const string & sep="" ) const;

		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );

	private :

		bool8 *	_values;

};



class Int8Values : public virtual Values
{
	public :

		// Constructors
		Int8Values(
			unsigned long			size,
			DataStorage *			parent );

		virtual ~Int8Values();

		int8 * int8Values( unsigned long from, unsigned long count );
		unsigned int dataSize() const;
		DataType dataType() const;
		void size( unsigned long size );
		void printOn( ostream& os, const string & sep="" ) const;

		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );

	private :

		int8 *	_values;

};


class Uint16Values : public virtual Values
{
	public :

		// Constructors
		Uint16Values(
			unsigned long			size,
			DataStorage *			parent );

		virtual ~Uint16Values();

		uint16 * uint16Values( unsigned long from, unsigned long count );
		unsigned int dataSize() const;
		DataType dataType() const;
		void size( unsigned long size );
		void printOn( ostream& os, const string & sep="" ) const;

		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );

	private :

		uint16 *	_values;

};


class Int16Values : public virtual Values
{
	public :

		// Constructors
		Int16Values(
			unsigned long			size,
			DataStorage *			parent );

		virtual ~Int16Values();

		int16 * int16Values( unsigned long from, unsigned long count );
		unsigned int dataSize() const;
		DataType dataType() const;
		void size( unsigned long size );
		void printOn( ostream& os, const string & sep="" ) const;

		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );

	private :

		int16 *	_values;

};


class Uint32Values : public virtual Values
{
	public :

		// Constructors
		Uint32Values(
			unsigned long			size,
			DataStorage *			parent );

		virtual ~Uint32Values();

		uint32 * uint32Values( unsigned long from, unsigned long count );
		unsigned int dataSize() const;
		DataType dataType() const;
		void size( unsigned long size );
		void printOn( ostream& os, const string & sep="" ) const;

		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );

	private :

		uint32 *	_values;
};


class Int32Values : public virtual Values
{
	public :

		// Constructors
		Int32Values(
			unsigned long			size,
			DataStorage *			parent );

		virtual ~Int32Values();

		int32 * int32Values( unsigned long from, unsigned long count );
		unsigned int dataSize() const;
		DataType dataType() const;
		void size( unsigned long size );
		void printOn( ostream& os, const string & sep="" ) const;

		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );

	private :

		int32 *	_values;

};


class Real32Values : public virtual Values
{
	public :

		// Constructors
		Real32Values(
			unsigned long			size,
			DataStorage *			parent );

		virtual ~Real32Values();

		real32 * real32Values( unsigned long from, unsigned long count );
		unsigned int dataSize() const;
		DataType dataType() const;
		void size( unsigned long size );
		void printOn( ostream& os, const string & sep="" ) const;

		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );

	private :

		real32 *	_values;

};


class Real64Values : public virtual Values
{
	public :

		// Constructors
		Real64Values(
			unsigned long			size,
			DataStorage *			parent );

		virtual ~Real64Values();

		real64 * real64Values( unsigned long from, unsigned long count );
		unsigned int dataSize() const;
		DataType dataType() const;
		void size( unsigned long size );
		void printOn( ostream& os, const string & sep="" ) const;

		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );

	private :

		real64 *	_values;

};


class StringValues : public virtual Values
{
	public :

		// Constructors
		StringValues(
			unsigned long			size,
			DataStorage *			parent );

		virtual ~StringValues();
		char * stringValues( unsigned long from, unsigned long count );
		unsigned int dataSize() const;
		DataType dataType() const;
		void size( unsigned long size );
		void printOn( ostream& os, const string & sep="" ) const;

		void insertBytes( unsigned long pos, unsigned long count );
		void deleteBytes( unsigned long from, unsigned long count );
		void copyBytes( unsigned long from, unsigned long to, unsigned long count );

	private :

		char *		_values;
};



// ****************************
// Values methods defined here.
// ****************************

Values::Values(
	unsigned long			size,
	DataStorage *			parent )
	:
		DalDal( parent -> mode(), parent ),
		DalChild<DataStorage>( parent ),
		DalSeekable<unsigned long>( 0, 0 ),
		_size( size )
{
	PROC
	ENDPROC
}


// ********************************
// BoolValues methods defined here.
// ********************************

BoolValues::BoolValues(
	unsigned long			size,
	DataStorage *			parent )
	:
		Values( size, parent ),
		_values( new bool8[ size ] )
{
	PROC
	ENDPROC
}


BoolValues::~BoolValues()
{
	PROC
	delete[] _values;
	ENDPROC
}


bool8 *
BoolValues::boolValues( unsigned long from, unsigned long /*count*/ )
{
	PROC
	RETURN _values + from;
}


unsigned int
BoolValues::dataSize() const
{
	PROC
	RETURN sizeof( bool8 );
}


Values::DataType
BoolValues::dataType() const
{
	PROC
	RETURN Bool;
}


void
BoolValues::size(
	unsigned long			size )
{
	PROC

	_size = size;

	delete[] _values;

	_values = new bool8[ size ];

	ENDPROC
}


void
BoolValues::printOn( ostream& os, const string & sep) const
{
	PROC
	writeValues( os, _values, _size, sep );
	ENDPROC

}

void BoolValues::insertBytes( unsigned long pos, unsigned long count)
{
	PROC
	::insertBytes( _values, _size, pos, count );
	ENDPROC
}


void BoolValues::deleteBytes( unsigned long from, unsigned long count)
{
	PROC
	::deleteBytes( _values, _size, from, count );
	ENDPROC
}


void BoolValues::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC
	::copyBytes( _values, _size, from, to, count );
	ENDPROC
}


// ********************************
// Int8Values methods defined here.
// ********************************

Int8Values::Int8Values(
	unsigned long			size,
	DataStorage *			parent )
	:
		Values( size, parent ),
		_values( new int8[ size ] )
{
	PROC
	ENDPROC
}


Int8Values::~Int8Values()
{
	PROC
	delete[] _values;
	ENDPROC
}


int8 *
Int8Values::int8Values( unsigned long from, unsigned long /*count*/ )
{
	PROC
	RETURN _values + from;
}


unsigned int
Int8Values::dataSize() const
{
	PROC
	RETURN sizeof( int8 );
}


Values::DataType
Int8Values::dataType() const
{
	PROC
	RETURN Int8;

}


void
Int8Values::size(
	unsigned long			size )
{
	PROC

	_size = size;

	delete[] _values;

	_values = new int8[ size ];

	ENDPROC
}


void
Int8Values::printOn( ostream& os, const string & sep) const
{
	PROC
	writeValues( os, _values, _size, sep );
	ENDPROC
}


void
Int8Values::insertBytes( unsigned long pos, unsigned long count)
{
	PROC

	::insertBytes( _values, _size, pos, count );

	ENDPROC
}


void
Int8Values::deleteBytes( unsigned long from, unsigned long count)
{
	PROC

	::deleteBytes( _values, _size, from, count );

	ENDPROC
}


void
Int8Values::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	::copyBytes( _values, _size, from, to, count );

	ENDPROC
}


// **********************************
// Uint16Values methods defined here.
// **********************************

Uint16Values::Uint16Values(
	unsigned long			size,
	DataStorage *			parent )
	:
		Values( size, parent ),
		_values( new uint16[ size ] )
{
	PROC
	ENDPROC
}


Uint16Values::~Uint16Values()
{
	PROC
	delete[] _values;
	ENDPROC
}


uint16 *
Uint16Values::uint16Values( unsigned long from, unsigned long /*count*/ )
{
	PROC
	RETURN _values + from;
}


unsigned int
Uint16Values::dataSize() const
{
	PROC
	RETURN sizeof( uint16 );
}


Values::DataType
Uint16Values::dataType() const
{
	PROC
	RETURN Uint16;
}


void
Uint16Values::size(
	unsigned long			size )
{
	PROC

	_size = size;

	delete[] _values;

	_values = new uint16[ size ];

	ENDPROC
}


void
Uint16Values::printOn( ostream& os, const string & sep) const
{
	writeValues( os, _values, _size, sep );

}


void
Uint16Values::insertBytes( unsigned long pos, unsigned long count)
{
	PROC

	::insertBytes( _values, _size, pos, count );

	ENDPROC
}


void
Uint16Values::deleteBytes( unsigned long from, unsigned long count)
{
	PROC

	::deleteBytes( _values, _size, from, count );

	ENDPROC
}


void Uint16Values::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	::copyBytes( _values, _size, from, to, count );

	ENDPROC
}


// *********************************
// Int16Values methods defined here.
// *********************************

Int16Values::Int16Values(
	unsigned long			size,
	DataStorage *			parent )
	:
		Values( size, parent ),
		_values( new int16[ size ] )
{
	PROC
	ENDPROC
}


Int16Values::~Int16Values()
{
	PROC
	delete[] _values;
	ENDPROC
}


int16 *
Int16Values::int16Values( unsigned long from, unsigned long /*count*/ )
{
	PROC
	RETURN _values + from;
}


unsigned int
Int16Values::dataSize() const
{
	PROC
	RETURN sizeof( int16 );
}


Values::DataType
Int16Values::dataType() const
{
	PROC
	RETURN Int16;
}


void
Int16Values::size(
	unsigned long			size )
{
	PROC

	_size = size;

	delete[] _values;

	_values = new int16[ size ];

	ENDPROC
}


void
Int16Values::printOn( ostream& os, const string & sep) const
{
	writeValues( os, _values, _size, sep );

}


void
Int16Values::insertBytes( unsigned long pos, unsigned long count)
{
	PROC

	::insertBytes( _values, _size, pos, count );

	ENDPROC
}


void
Int16Values::deleteBytes( unsigned long from, unsigned long count)
{
	PROC

	::deleteBytes( _values, _size, from, count );

	ENDPROC
}


void Int16Values::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	::copyBytes( _values, _size, from, to, count );

	ENDPROC
}


// **********************************
// Uint32Values methods defined here.
// **********************************

Uint32Values::Uint32Values(
	unsigned long			size,
	DataStorage *			parent )
	:
		Values( size, parent ),
		_values( new uint32[ size ] )
{
	PROC
	ENDPROC
}


Uint32Values::~Uint32Values()
{
	PROC
	delete[] _values;
	ENDPROC
}


uint32 *
Uint32Values::uint32Values( unsigned long from, unsigned long /*count*/ )
{
	PROC
	RETURN _values + from;

}


unsigned int
Uint32Values::dataSize() const
{
	PROC
	RETURN sizeof( uint32 );
}


Values::DataType
Uint32Values::dataType() const
{
	PROC
	RETURN Uint32;
}


void
Uint32Values::size(
	unsigned long			size )
{
	PROC

	_size = size;

	delete[] _values;

	_values = new uint32[ size ];

	ENDPROC
}


void
Uint32Values::printOn( ostream& os, const string & sep) const
{
	writeValues( os, _values, _size, sep );

}


void
Uint32Values::insertBytes( unsigned long pos, unsigned long count)
{
	PROC

	::insertBytes( _values, _size, pos, count );

	ENDPROC
}


void
Uint32Values::deleteBytes( unsigned long from, unsigned long count)
{
	PROC

	::deleteBytes( _values, _size, from, count );

	ENDPROC
}


void
Uint32Values::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	::copyBytes( _values, _size, from, to, count );

	ENDPROC
}


// *********************************
// Int32Values methods defined here.
// *********************************

Int32Values::Int32Values(
	unsigned long			size,
	DataStorage *			parent )
	:
		Values( size, parent ),
		_values( new int32[ size ] )
{
	PROC
	ENDPROC
}


Int32Values::~Int32Values()
{
	PROC

	delete[] _values;

	ENDPROC
}


int32 *
Int32Values::int32Values( unsigned long from, unsigned long /*count*/ )
{
	PROC
	RETURN _values + from;

}


unsigned int
Int32Values::dataSize() const
{
	PROC
	RETURN sizeof( int32 );
}


Values::DataType
Int32Values::dataType() const
{
	PROC
	RETURN Int32;
}


void
Int32Values::size(
	unsigned long			size )
{
	PROC

	_size = size;

	delete[] _values;

	_values = new int32[ size ];

	ENDPROC
}


void
Int32Values::printOn( ostream& os, const string & sep) const
{
	writeValues( os, _values, _size, sep );

}


void
Int32Values::insertBytes( unsigned long pos, unsigned long count)
{
	PROC

	::insertBytes( _values, _size, pos, count );

	ENDPROC
}


void
Int32Values::deleteBytes( unsigned long from, unsigned long count)
{
	PROC

	::deleteBytes( _values, _size, from, count );

	ENDPROC
}


void
Int32Values::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	::copyBytes( _values, _size, from, to, count );

	ENDPROC
}


// **********************************
// Real32Values methods defined here.
// **********************************

Real32Values::Real32Values(
	unsigned long			size,
	DataStorage *			parent )
	:
		Values( size, parent ),
		_values( new real32[ size ] )
{
	PROC
	ENDPROC
}


Real32Values::~Real32Values()
{
	PROC
	delete[] _values;
	ENDPROC
}


real32 *
Real32Values::real32Values( unsigned long from, unsigned long /*count*/ )
{
	PROC
	RETURN _values + from;
}


unsigned int
Real32Values::dataSize() const
{
	PROC
	RETURN sizeof( real32 );
}


Values::DataType
Real32Values::dataType() const
{
	PROC
	RETURN Real32;
}


void
Real32Values::size(
	unsigned long			size )
{
	PROC

	_size = size;

	delete[] _values;

	_values = new real32[ size ];

	ENDPROC
}


void
Real32Values::printOn( ostream& os, const string & sep) const
{
	writeValues( os, _values, _size, sep );

}


void
Real32Values::insertBytes( unsigned long pos, unsigned long count)
{
	PROC

	::insertBytes( _values, _size, pos, count );

	ENDPROC
}


void
Real32Values::deleteBytes( unsigned long from, unsigned long count)
{
	PROC

	::deleteBytes( _values, _size, from, count );

	ENDPROC
}


void
Real32Values::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	::copyBytes( _values, _size, from, to, count );

	ENDPROC
}



// **********************************
// Real64Values methods defined here.
// **********************************

Real64Values::Real64Values(
	unsigned long			size,
	DataStorage *			parent )
	:
		Values( size, parent ),
		_values( new real64[ size ] )
{
	PROC
	ENDPROC
}


Real64Values::~Real64Values()
{
	PROC
	delete[] _values;
	ENDPROC
}


real64 *
Real64Values::real64Values( unsigned long from, unsigned long /*count*/ )
{
	PROC
	RETURN _values + from;
}


unsigned int
Real64Values::dataSize() const
{
	PROC
	RETURN sizeof( real64 );
}


Values::DataType
Real64Values::dataType() const
{
	PROC
	RETURN Real64;
}


void
Real64Values::size(
	unsigned long			size )
{
	PROC

	_size = size;

	delete[] _values;

	_values = new real64[ size ];

	ENDPROC
}


void
Real64Values::printOn( ostream& os, const string & sep) const
{
	writeValues( os, _values, _size, sep );

}


void
Real64Values::insertBytes( unsigned long pos, unsigned long count)
{
	PROC

	::insertBytes( _values, _size, pos, count );

	ENDPROC
}


void
Real64Values::deleteBytes( unsigned long from, unsigned long count)
{
	PROC

	::deleteBytes( _values, _size, from, count );

	ENDPROC
}


void
Real64Values::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	::copyBytes( _values, _size, from, to, count );

	ENDPROC
}


// **********************************
// StringValues methods defined here.
// **********************************

StringValues::StringValues(
	unsigned long			size,
	DataStorage *			parent )
	:
		Values( size, parent ),
		_values( new char[ size ] )
{
	PROC
	ENDPROC
}


StringValues::~StringValues()
{
	PROC
	delete[] _values;
	ENDPROC
}


char *
StringValues::stringValues( unsigned long from, unsigned long /*count*/ )
{
	PROC
	RETURN _values + from;
}


unsigned int
StringValues::dataSize() const
{
	PROC
	RETURN sizeof( dstring );
}


Values::DataType
StringValues::dataType() const
{
	PROC
	RETURN DString;
}


void
StringValues::size(
	unsigned long			size )
{
	PROC

	_size = size;

	delete[] _values;

	_values = new char[ size ];

	ENDPROC
}


void
StringValues::printOn( ostream& os, const string & sep) const
{
	writeValues( os, _values, _size, sep );

}


void
StringValues::insertBytes( unsigned long pos, unsigned long count)
{
	PROC

	::insertBytes( _values, _size, pos, count );

	ENDPROC
}


void
StringValues::deleteBytes( unsigned long from, unsigned long count)
{
	PROC

	::deleteBytes( _values, _size, from, count );

	ENDPROC
}


void
StringValues::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	::copyBytes( _values, _size, from, to, count );

	ENDPROC
}


// *********************************
// DataStorage methods defined here.
// *********************************

DataStorage::DataStorage(
			const string&			name,
			DataType			dataType,
			unsigned long			size )
	:
		DalDal( Dal::Read, 0 ),
		_name(name)
{
	PROC

	SAS_DEBUG( clog << "Constructing DataStorage name=" << name << " type=" << int( dataType ) << " size=" << size << endl; )

	switch( dataType )
	{
		case Bool :
			_values = new BoolValues( size, this );
			break;
		case Int8 :
			_values = new Int8Values( size, this );
			break;
		case Uint16 :
			_values = new Uint16Values( size, this );
			break;
		case Int16 :
			_values = new Int16Values( size, this );
			break;
		case Uint32 :
			_values = new Uint32Values( size, this );
			break;
		case Int32 :
			_values = new Int32Values( size, this );
			break;
		case Real32 :
			_values = new Real32Values( size, this );
			break;
		case Real64 :
			_values = new Real64Values( size, this );
			break;
		case DString :
			_values = new StringValues( size, this );
			break;
		default :
			errstr << "encountered invalid type " << int( dataType)<< " for data storage of " << _name << Sas::error("internalError");
			break;
	} // switch

	ENDPROC
}



DataStorage::~DataStorage()
{
	PROC

	delete _values;

	ENDPROC
}


void
DataStorage::cannotConvertTo(const string& type) const
{
	errstr << "Accessing data " << _name << " of type " << toString( _values -> dataType() ) << " as " << type << " is not supported" << Sas::error("accessingAsWrongDataType");
}


int8 *
DataStorage::int8Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Int8 )
	{
		// The following code used to convert the data
		// to the type asked. This is abandoned, since
		// the old data type is irrevocably lost and
		// the new data type will be written to the output.
		//
		//Values * values = new Int8Values(
		//			_values -> rows(),
		//			_values -> size() );
		//
		//_values -> getValues( values -> int8Values() );
		//delete _values;
		//_values = values;

		//BUG: should report name of object.
		//errstr << "Accessing data " << _name << " of type " << int( _values -> dataType() ) << " as int8 is not supported" << Sas::error("accessingAsWrongDataType");
		cannotConvertTo("int8");

	} // if

	RETURN _values -> int8Values( from, count );

}


uint16 *
DataStorage::uint16Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Uint16 )
	{
		cannotConvertTo("uint16");
	} // if

	RETURN _values -> uint16Values( from, count );

}


int16 *
DataStorage::int16Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Int16 )
	{
		cannotConvertTo("int16");
	} // if

	RETURN _values -> int16Values( from, count );

}


uint32 *
DataStorage::uint32Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Uint32 )
	{
		cannotConvertTo("uint32");
	} // if

	RETURN _values -> uint32Values( from, count );

}


int32 *
DataStorage::int32Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Int32 )
	{
		cannotConvertTo("int32");
	} // if

	RETURN _values -> int32Values( from, count );

}


real32 *
DataStorage::real32Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Real32 )
	{
		cannotConvertTo("real32");
	} // if

	RETURN _values -> real32Values( from, count );

}


real64 *
DataStorage::real64Data( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Real64 )
	{
		cannotConvertTo("real64");
	} // if

	RETURN _values -> real64Values( from, count );

}


bool8 *
DataStorage::boolData( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != Bool )
	{
		// *********************************
		// Convert Data to appropriate type.
		// *********************************

		// ****************
		// NOT IMPLEMENTED.
		// ****************
		//BUG: should report name of object.
		cannotConvertTo("string");
	} // if

	RETURN _values -> boolValues( from, count );

}


char *
DataStorage::stringData( unsigned long from, unsigned long count )
{
	PROC

	if( _values -> dataType() != DString )
	{
		cannotConvertTo("string");
	} // if

	RETURN _values -> stringValues( from, count );

}


unsigned int
DataStorage::dataSize() const
{
	PROC RETURN _values -> dataSize();

}


DataStorage::DataType
DataStorage::dataType() const
{
	PROC
	RETURN _values -> dataType();

}


unsigned long
DataStorage::size() const
{
	PROC
	RETURN _values -> size();
}


void
DataStorage::size(unsigned long size )
{
	PROC
	_values -> size( size );
	ENDPROC
}


unsigned long
DataStorage::from() const
{
	PROC
	RETURN _values -> from();
}


unsigned long
DataStorage::count() const
{
	PROC
	RETURN _values -> count();
}


void
DataStorage::insertBytes( unsigned long pos, unsigned long count )
{
	PROC

	_values -> insertBytes( pos, count );

	ENDPROC
}


void
DataStorage::deleteBytes( unsigned long from, unsigned long count )
{
	PROC

	_values -> deleteBytes( from, count );

	ENDPROC
}


void
DataStorage::copyBytes( unsigned long from, unsigned long to, unsigned long count )
{
	PROC

	_values -> copyBytes( from, to, count );

	ENDPROC
}


void
DataStorage::printOn( ostream & os, const string & sep)
{
	PROC

	_values -> printOn( os, sep );

	ENDPROC
}


void
DataStorage::release()
{
	PROC
	ENDPROC
}


